from collections.abc import Iterable
from typing import Literal

import numpy as np
import torch
from torch import nn
from torch.distributions import Normal, Poisson
from torch.distributions import kl_divergence as kld
from torch.nn import functional as F

from scvi import REGISTRY_KEYS
from scvi.distributions import (
    NegativeBinomial,
    NegativeBinomialMixture,
    ZeroInflatedNegativeBinomial,
)
from scvi.module._peakvae import Decoder as DecoderPeakVI
from scvi.module.base import BaseModuleClass, LossOutput, auto_move_data
from scvi.nn import DecoderSCVI, Encoder, FCLayers

from ._utils import masked_softmax


class LibrarySizeEncoder(torch.nn.Module):
    """Library size encoder."""

    def __init__(
        self,
        n_input: int,
        n_cat_list: Iterable[int] = None,
        n_layers: int = 2,
        n_hidden: int = 128,
        use_batch_norm: bool = False,
        use_layer_norm: bool = True,
        deep_inject_covariates: bool = False,
        **kwargs,
    ):
        super().__init__()
        self.px_decoder = FCLayers(
            n_in=n_input,
            n_out=n_hidden,
            n_cat_list=n_cat_list,
            n_layers=n_layers,
            n_hidden=n_hidden,
            dropout_rate=0,
            activation_fn=torch.nn.LeakyReLU,
            use_batch_norm=use_batch_norm,
            use_layer_norm=use_layer_norm,
            inject_covariates=deep_inject_covariates,
            **kwargs,
        )
        self.output = torch.nn.Sequential(torch.nn.Linear(n_hidden, 1), torch.nn.LeakyReLU())

    def forward(self, x: torch.Tensor, *cat_list: int):
        """Forward pass."""
        return self.output(self.px_decoder(x, *cat_list))


class DecoderADT(torch.nn.Module):
    """Decoder for just surface proteins (ADT)."""

    def __init__(
        self,
        n_input: int,
        n_output_proteins: int,
        n_cat_list: Iterable[int] = None,
        n_layers: int = 2,
        n_hidden: int = 128,
        dropout_rate: float = 0.1,
        use_batch_norm: bool = False,
        use_layer_norm: bool = True,
        deep_inject_covariates: bool = False,
    ):
        super().__init__()
        self.n_output_proteins = n_output_proteins

        linear_args = {
            "n_layers": 1,
            "use_activation": False,
            "use_batch_norm": False,
            "use_layer_norm": False,
            "dropout_rate": 0,
        }

        self.py_fore_decoder = FCLayers(
            n_in=n_input,
            n_out=n_hidden,
            n_cat_list=n_cat_list,
            n_layers=n_layers,
            n_hidden=n_hidden,
            dropout_rate=dropout_rate,
            use_batch_norm=use_batch_norm,
            use_layer_norm=use_layer_norm,
        )
        self.py_fore_scale_decoder = FCLayers(
            n_in=n_hidden + n_input,
            n_out=n_output_proteins,
            n_cat_list=n_cat_list,
            n_layers=1,
            use_activation=True,
            use_batch_norm=False,
            use_layer_norm=False,
            dropout_rate=0,
            activation_fn=nn.ReLU,
        )

        self.py_background_decoder = FCLayers(
            n_in=n_hidden + n_input,
            n_out=n_output_proteins,
            n_cat_list=n_cat_list,
            **linear_args,
        )

        # dropout (mixture component for proteins, ZI probability for genes)
        self.sigmoid_decoder = FCLayers(
            n_in=n_input,
            n_out=n_hidden,
            n_cat_list=n_cat_list,
            n_layers=n_layers,
            n_hidden=n_hidden,
            dropout_rate=dropout_rate,
            use_batch_norm=use_batch_norm,
            use_layer_norm=use_layer_norm,
        )

        # background mean parameters second decoder
        self.py_back_mean_log_alpha = FCLayers(
            n_in=n_hidden + n_input,
            n_out=n_output_proteins,
            n_cat_list=n_cat_list,
            **linear_args,
        )
        self.py_back_mean_log_beta = FCLayers(
            n_in=n_hidden + n_input,
            n_out=n_output_proteins,
            n_cat_list=n_cat_list,
            **linear_args,
        )

        # background mean first decoder
        self.py_back_decoder = FCLayers(
            n_in=n_input,
            n_out=n_hidden,
            n_cat_list=n_cat_list,
            n_layers=n_layers,
            n_hidden=n_hidden,
            dropout_rate=dropout_rate,
            use_batch_norm=use_batch_norm,
            use_layer_norm=use_layer_norm,
        )

    def forward(self, z: torch.Tensor, *cat_list: int):
        """Forward pass."""
        # z is the latent repr
        py_ = {}

        py_back = self.py_back_decoder(z, *cat_list)
        py_back_cat_z = torch.cat([py_back, z], dim=-1)

        py_["back_alpha"] = self.py_back_mean_log_alpha(py_back_cat_z, *cat_list)
        py_["back_beta"] = torch.exp(self.py_back_mean_log_beta(py_back_cat_z, *cat_list))
        log_pro_back_mean = Normal(py_["back_alpha"], py_["back_beta"]).rsample()
        py_["rate_back"] = torch.exp(log_pro_back_mean)

        py_fore = self.py_fore_decoder(z, *cat_list)
        py_fore_cat_z = torch.cat([py_fore, z], dim=-1)
        py_["fore_scale"] = self.py_fore_scale_decoder(py_fore_cat_z, *cat_list) + 1 + 1e-8
        py_["rate_fore"] = py_["rate_back"] * py_["fore_scale"]

        p_mixing = self.sigmoid_decoder(z, *cat_list)
        p_mixing_cat_z = torch.cat([p_mixing, z], dim=-1)
        py_["mixing"] = self.py_background_decoder(p_mixing_cat_z, *cat_list)

        protein_mixing = 1 / (1 + torch.exp(-py_["mixing"]))
        py_["scale"] = torch.nn.functional.normalize(
            (1 - protein_mixing) * py_["rate_fore"], p=1, dim=-1
        )

        return py_, log_pro_back_mean


class MULTIVAE(BaseModuleClass):
    """Variational auto-encoder model for joint paired + unpaired RNA-seq and ATAC-seq data.

    Parameters
    ----------
    n_input_regions
        Number of input regions.
    n_input_genes
        Number of input genes.
    n_input_proteins
        Number of input proteins
    modality_weights
        Weighting scheme across modalities. One of the following:
        * ``"equal"``: Equal weight in each modality
        * ``"universal"``: Learn weights across modalities w_m.
        * ``"cell"``: Learn weights across modalities and cells. w_{m,c}
    modality_penalty
        Training Penalty across modalities. One of the following:
        * ``"Jeffreys"``: Jeffreys penalty to align modalities
        * ``"MMD"``: MMD penalty to align modalities
        * ``"None"``: No penalty
    n_batch
        Number of batches, if 0, no batch correction is performed.
    gene_likelihood
        The distribution to use for gene expression data. One of the following
        * ``'zinb'`` - Zero-Inflated Negative Binomial
        * ``'nb'`` - Negative Binomial
        * ``'poisson'`` - Poisson
    gene_dispersion
        One of the following:
        * ``'gene'`` - dispersion parameter of NB is constant per gene across cells
        * ``'gene-batch'`` - dispersion can differ between different batches
        * ``'gene-label'`` - dispersion can differ between different labels
        * ``'gene-cell'`` - dispersion can differ for every gene in every cell
    protein_dispersion
        One of the following:

        * ``'protein'`` - protein_dispersion parameter is constant per protein across cells
        * ``'protein-batch'`` - protein_dispersion can differ between different batches NOT TESTED
        * ``'protein-label'`` - protein_dispersion can differ between different labels NOT TESTED
    n_hidden
        Number of nodes per hidden layer. If `None`, defaults to square root
        of number of regions.
    n_latent
        Dimensionality of the latent space. If `None`, defaults to square root
        of `n_hidden`.
    n_layers_encoder
        Number of hidden layers used for encoder NN.
    n_layers_decoder
        Number of hidden layers used for decoder NN.
    dropout_rate
        Dropout rate for neural networks
    region_factors
        Include region-specific factors in the model
    use_batch_norm
        One of the following
        * ``'encoder'`` - use batch normalization in the encoder only
        * ``'decoder'`` - use batch normalization in the decoder only
        * ``'none'`` - do not use batch normalization
        * ``'both'`` - use batch normalization in both the encoder and decoder
    use_layer_norm
        One of the following
        * ``'encoder'`` - use layer normalization in the encoder only
        * ``'decoder'`` - use layer normalization in the decoder only
        * ``'none'`` - do not use layer normalization
        * ``'both'`` - use layer normalization in both the encoder and decoder
    latent_distribution
        which latent distribution to use, options are
        * ``'normal'`` - Normal distribution
        * ``'ln'`` - Logistic normal distribution (Normal(0, I) transformed by softmax)
    deeply_inject_covariates
        Whether to deeply inject covariates into all layers of the decoder. If False,
        covariates will only be included in the input layer.
    encode_covariates
        If True, include covariates in the input to the encoder.
    use_size_factor_key
        Use size_factor AnnDataField defined by the user as scaling factor in mean of conditional
        RNA distribution.
    """

    def __init__(
        self,
        n_input_regions: int = 0,
        n_input_genes: int = 0,
        n_input_proteins: int = 0,
        modality_weights: Literal["equal", "cell", "universal"] = "equal",
        modality_penalty: Literal["Jeffreys", "MMD", "None"] = "Jeffreys",
        n_batch: int = 0,
        n_obs: int = 0,
        n_labels: int = 0,
        gene_likelihood: Literal["zinb", "nb", "poisson"] = "zinb",
        gene_dispersion: Literal["gene", "gene-batch", "gene-label", "gene-cell"] = "gene",
        n_hidden: int = None,
        n_latent: int = None,
        n_layers_encoder: int = 2,
        n_layers_decoder: int = 2,
        n_continuous_cov: int = 0,
        n_cats_per_cov: Iterable[int] | None = None,
        dropout_rate: float = 0.1,
        region_factors: bool = True,
        use_batch_norm: Literal["encoder", "decoder", "none", "both"] = "none",
        use_layer_norm: Literal["encoder", "decoder", "none", "both"] = "both",
        latent_distribution: Literal["normal", "ln"] = "normal",
        deeply_inject_covariates: bool = False,
        encode_covariates: bool = False,
        use_size_factor_key: bool = False,
        protein_background_prior_mean: np.ndarray | None = None,
        protein_background_prior_scale: np.ndarray | None = None,
        protein_dispersion: str = "protein",
    ):
        super().__init__()

        # INIT PARAMS
        self.n_input_regions = n_input_regions
        self.n_input_genes = n_input_genes
        self.n_input_proteins = n_input_proteins
        if n_hidden is None:
            if n_input_regions == 0:
                self.n_hidden = np.min([128, int(np.sqrt(self.n_input_genes))])
            else:
                self.n_hidden = np.min([128, int(np.sqrt(self.n_input_regions))])
        else:
            self.n_hidden = n_hidden
        self.n_batch = n_batch

        self.gene_likelihood = gene_likelihood
        self.latent_distribution = latent_distribution

        self.n_latent = int(np.sqrt(self.n_hidden)) if n_latent is None else n_latent
        self.n_layers_encoder = n_layers_encoder
        self.n_layers_decoder = n_layers_decoder
        self.n_cats_per_cov = n_cats_per_cov
        self.n_continuous_cov = n_continuous_cov
        self.dropout_rate = dropout_rate

        self.use_batch_norm_encoder = use_batch_norm in ("encoder", "both")
        self.use_batch_norm_decoder = use_batch_norm in ("decoder", "both")
        self.use_layer_norm_encoder = use_layer_norm in ("encoder", "both")
        self.use_layer_norm_decoder = use_layer_norm in ("decoder", "both")
        self.encode_covariates = encode_covariates
        self.deeply_inject_covariates = deeply_inject_covariates
        self.use_size_factor_key = use_size_factor_key

        cat_list = [n_batch] + list(n_cats_per_cov) if n_cats_per_cov is not None else []
        encoder_cat_list = cat_list if encode_covariates else None

        # expression
        # expression dispersion parameters
        self.gene_likelihood = gene_likelihood
        self.gene_dispersion = gene_dispersion
        if self.gene_dispersion == "gene":
            self.px_r = torch.nn.Parameter(torch.randn(n_input_genes))
        elif self.gene_dispersion == "gene-batch":
            self.px_r = torch.nn.Parameter(torch.randn(n_input_genes, n_batch))
        elif self.gene_dispersion == "gene-label":
            self.px_r = torch.nn.Parameter(torch.randn(n_input_genes, n_labels))
        elif self.gene_dispersion == "gene-cell":
            pass
        else:
            raise ValueError(
                "dispersion must be one of ['gene', 'gene-batch',"
                " 'gene-label', 'gene-cell'], but input was "
                "{}.format(self.dispersion)"
            )

        # expression encoder
        if self.n_input_genes == 0:
            input_exp = 1
        else:
            input_exp = self.n_input_genes
        n_input_encoder_exp = input_exp + n_continuous_cov * encode_covariates
        self.z_encoder_expression = Encoder(
            n_input=n_input_encoder_exp,
            n_output=self.n_latent,
            n_cat_list=encoder_cat_list,
            n_layers=self.n_layers_encoder,
            n_hidden=self.n_hidden,
            dropout_rate=self.dropout_rate,
            distribution=self.latent_distribution,
            inject_covariates=deeply_inject_covariates,
            use_batch_norm=self.use_batch_norm_encoder,
            use_layer_norm=self.use_layer_norm_encoder,
            activation_fn=torch.nn.LeakyReLU,
            var_eps=0,
            return_dist=False,
        )

        # expression library size encoder
        self.l_encoder_expression = LibrarySizeEncoder(
            n_input_encoder_exp,
            n_cat_list=encoder_cat_list,
            n_layers=self.n_layers_encoder,
            n_hidden=self.n_hidden,
            use_batch_norm=self.use_batch_norm_encoder,
            use_layer_norm=self.use_layer_norm_encoder,
            deep_inject_covariates=self.deeply_inject_covariates,
        )

        # expression decoder
        n_input_decoder = self.n_latent + self.n_continuous_cov
        self.z_decoder_expression = DecoderSCVI(
            n_input_decoder,
            n_input_genes,
            n_cat_list=cat_list,
            n_layers=n_layers_decoder,
            n_hidden=self.n_hidden,
            inject_covariates=self.deeply_inject_covariates,
            use_batch_norm=self.use_batch_norm_decoder,
            use_layer_norm=self.use_layer_norm_decoder,
            scale_activation="softplus" if use_size_factor_key else "softmax",
        )

        # accessibility
        # accessibility encoder
        if self.n_input_regions == 0:
            input_acc = 1
        else:
            input_acc = self.n_input_regions
        n_input_encoder_acc = input_acc + n_continuous_cov * encode_covariates
        self.z_encoder_accessibility = Encoder(
            n_input=n_input_encoder_acc,
            n_layers=self.n_layers_encoder,
            n_output=self.n_latent,
            n_hidden=self.n_hidden,
            n_cat_list=encoder_cat_list,
            dropout_rate=self.dropout_rate,
            activation_fn=torch.nn.LeakyReLU,
            distribution=self.latent_distribution,
            var_eps=0,
            use_batch_norm=self.use_batch_norm_encoder,
            use_layer_norm=self.use_layer_norm_encoder,
            return_dist=False,
        )

        # accessibility region-specific factors
        self.region_factors = None
        if region_factors:
            self.region_factors = torch.nn.Parameter(torch.zeros(self.n_input_regions))

        # accessibility decoder
        self.z_decoder_accessibility = DecoderPeakVI(
            n_input=self.n_latent + self.n_continuous_cov,
            n_output=n_input_regions,
            n_hidden=self.n_hidden,
            n_cat_list=cat_list,
            n_layers=self.n_layers_decoder,
            use_batch_norm=self.use_batch_norm_decoder,
            use_layer_norm=self.use_layer_norm_decoder,
            deep_inject_covariates=self.deeply_inject_covariates,
        )

        # accessibility library size encoder
        self.l_encoder_accessibility = DecoderPeakVI(
            n_input=n_input_encoder_acc,
            n_output=1,
            n_hidden=self.n_hidden,
            n_cat_list=encoder_cat_list,
            n_layers=self.n_layers_encoder,
            use_batch_norm=self.use_batch_norm_encoder,
            use_layer_norm=self.use_layer_norm_encoder,
            deep_inject_covariates=self.deeply_inject_covariates,
        )

        # protein
        # protein encoder
        self.protein_dispersion = protein_dispersion
        if protein_background_prior_mean is None:
            if n_batch > 0:
                self.background_pro_alpha = torch.nn.Parameter(
                    torch.randn(n_input_proteins, n_batch)
                )
                self.background_pro_log_beta = torch.nn.Parameter(
                    torch.clamp(torch.randn(n_input_proteins, n_batch), -10, 1)
                )
            else:
                self.background_pro_alpha = torch.nn.Parameter(torch.randn(n_input_proteins))
                self.background_pro_log_beta = torch.nn.Parameter(
                    torch.clamp(torch.randn(n_input_proteins), -10, 1)
                )
        else:
            if protein_background_prior_mean.shape[1] == 1 and n_batch != 1:
                init_mean = protein_background_prior_mean.ravel()
                init_scale = protein_background_prior_scale.ravel()
            else:
                init_mean = protein_background_prior_mean
                init_scale = protein_background_prior_scale
            self.background_pro_alpha = torch.nn.Parameter(
                torch.from_numpy(init_mean.astype(np.float32))
            )
            self.background_pro_log_beta = torch.nn.Parameter(
                torch.log(torch.from_numpy(init_scale.astype(np.float32)))
            )

        # protein encoder
        if self.n_input_proteins == 0:
            input_pro = 1
        else:
            input_pro = self.n_input_proteins
        n_input_encoder_pro = input_pro + n_continuous_cov * encode_covariates
        self.z_encoder_protein = Encoder(
            n_input=n_input_encoder_pro,
            n_layers=self.n_layers_encoder,
            n_output=self.n_latent,
            n_hidden=self.n_hidden,
            n_cat_list=encoder_cat_list,
            dropout_rate=self.dropout_rate,
            activation_fn=torch.nn.LeakyReLU,
            distribution=self.latent_distribution,
            var_eps=0,
            use_batch_norm=self.use_batch_norm_encoder,
            use_layer_norm=self.use_layer_norm_encoder,
            return_dist=False,
        )

        # protein decoder
        self.z_decoder_pro = DecoderADT(
            n_input=n_input_decoder,
            n_output_proteins=n_input_proteins,
            n_hidden=self.n_hidden,
            n_cat_list=cat_list,
            n_layers=self.n_layers_decoder,
            use_batch_norm=self.use_batch_norm_decoder,
            use_layer_norm=self.use_layer_norm_decoder,
            deep_inject_covariates=self.deeply_inject_covariates,
        )

        # protein dispersion parameters
        if self.protein_dispersion == "protein":
            self.py_r = torch.nn.Parameter(2 * torch.rand(self.n_input_proteins))
        elif self.protein_dispersion == "protein-batch":
            self.py_r = torch.nn.Parameter(2 * torch.rand(self.n_input_proteins, n_batch))
        elif self.protein_dispersion == "protein-label":
            self.py_r = torch.nn.Parameter(2 * torch.rand(self.n_input_proteins, n_labels))
        else:  # protein-cell
            pass

        # modality alignment
        self.n_obs = n_obs
        self.modality_weights = modality_weights
        self.modality_penalty = modality_penalty
        self.n_modalities = int(n_input_genes > 0) + int(n_input_regions > 0)
        max_n_modalities = 3
        if modality_weights == "equal":
            mod_weights = torch.ones(max_n_modalities)
            self.register_buffer("mod_weights", mod_weights)
        elif modality_weights == "universal":
            self.mod_weights = torch.nn.Parameter(torch.ones(max_n_modalities))
        else:  # cell-specific weights
            self.mod_weights = torch.nn.Parameter(torch.ones(n_obs, max_n_modalities))

    def _get_inference_input(self, tensors):
        """Get input tensors for the inference model."""
        x = tensors.get(REGISTRY_KEYS.X_KEY, None)
        x_atac = tensors.get(REGISTRY_KEYS.ATAC_X_KEY, None)
        if x is not None and x_atac is not None:
            x = torch.cat((x, x_atac), dim=-1)
        elif x is None:
            x = x_atac
        if self.n_input_proteins == 0:
            y = torch.zeros(x.shape[0], 1, device=x.device, requires_grad=False)
        else:
            y = tensors[REGISTRY_KEYS.PROTEIN_EXP_KEY]
        batch_index = tensors[REGISTRY_KEYS.BATCH_KEY]
        cell_idx = tensors.get(REGISTRY_KEYS.INDICES_KEY).long().ravel()
        cont_covs = tensors.get(REGISTRY_KEYS.CONT_COVS_KEY)
        cat_covs = tensors.get(REGISTRY_KEYS.CAT_COVS_KEY)
        label = tensors[REGISTRY_KEYS.LABELS_KEY]
        size_factor = tensors.get(REGISTRY_KEYS.SIZE_FACTOR_KEY, None)
        input_dict = {
            "x": x,
            "y": y,
            "batch_index": batch_index,
            "cont_covs": cont_covs,
            "cat_covs": cat_covs,
            "label": label,
            "cell_idx": cell_idx,
            "size_factor": size_factor,
        }
        return input_dict

    @auto_move_data
    def inference(
        self,
        x,
        y,
        batch_index,
        cont_covs,
        cat_covs,
        label,
        cell_idx,
        size_factor,
        n_samples=1,
    ) -> dict[str, torch.Tensor]:
        """Run the inference model."""
        # Get Data and Additional Covs
        if self.n_input_genes == 0:
            x_rna = torch.zeros(x.shape[0], 1, device=x.device, requires_grad=False)
        else:
            x_rna = x[:, : self.n_input_genes]
        if self.n_input_regions == 0:
            x_atac = torch.zeros(x.shape[0], 1, device=x.device, requires_grad=False)
        else:
            x_atac = x[:, self.n_input_genes : (self.n_input_genes + self.n_input_regions)]

        mask_expr = x_rna.sum(dim=1) > 0
        mask_acc = x_atac.sum(dim=1) > 0
        mask_pro = y.sum(dim=1) > 0

        if cont_covs is not None and self.encode_covariates:
            encoder_input_expression = torch.cat((x_rna, cont_covs), dim=-1)
            encoder_input_accessibility = torch.cat((x_atac, cont_covs), dim=-1)
            encoder_input_protein = torch.cat((y, cont_covs), dim=-1)
        else:
            encoder_input_expression = x_rna
            encoder_input_accessibility = x_atac
            encoder_input_protein = y

        if cat_covs is not None and self.encode_covariates:
            categorical_input = torch.split(cat_covs, 1, dim=1)
        else:
            categorical_input = ()

        # Z Encoders
        qzm_acc, qzv_acc, z_acc = self.z_encoder_accessibility(
            encoder_input_accessibility, batch_index, *categorical_input
        )
        qzm_expr, qzv_expr, z_expr = self.z_encoder_expression(
            encoder_input_expression, batch_index, *categorical_input
        )
        qzm_pro, qzv_pro, z_pro = self.z_encoder_protein(
            encoder_input_protein, batch_index, *categorical_input
        )

        # L encoders
        if self.use_size_factor_key:
            libsize_expr = torch.log(size_factor[:, [0]] + 1e-6)
            libsize_acc = size_factor[:, [1]]
        else:
            libsize_expr = self.l_encoder_expression(
                encoder_input_expression, batch_index, *categorical_input
            )
            libsize_acc = self.l_encoder_accessibility(
                encoder_input_accessibility, batch_index, *categorical_input
            )

        # mix representations
        if self.modality_weights == "cell":
            weights = self.mod_weights[cell_idx, :]
        else:
            weights = self.mod_weights.unsqueeze(0).expand(len(cell_idx), -1)

        qz_m = mix_modalities(
            (qzm_expr, qzm_acc, qzm_pro), (mask_expr, mask_acc, mask_pro), weights
        )
        qz_v = mix_modalities(
            (qzv_expr, qzv_acc, qzv_pro),
            (mask_expr, mask_acc, mask_pro),
            weights,
            torch.sqrt,
        )

        # sample
        if n_samples > 1:

            def unsqz(zt, n_s):
                return zt.unsqueeze(0).expand((n_s, zt.size(0), zt.size(1)))

            untran_za = Normal(qzm_acc, qzv_acc.sqrt()).sample((n_samples,))
            z_acc = self.z_encoder_accessibility.z_transformation(untran_za)
            untran_ze = Normal(qzm_expr, qzv_expr.sqrt()).sample((n_samples,))
            z_expr = self.z_encoder_expression.z_transformation(untran_ze)
            untran_zp = Normal(qzm_pro, qzv_pro.sqrt()).sample((n_samples,))
            z_pro = self.z_encoder_protein.z_transformation(untran_zp)

            libsize_expr = unsqz(libsize_expr, n_samples)
            libsize_acc = unsqz(libsize_acc, n_samples)

        # sample from the mixed representation
        untran_z = Normal(qz_m, qz_v.sqrt()).rsample()
        z = self.z_encoder_accessibility.z_transformation(untran_z)

        outputs = {
            "x": x,
            "z": z,
            "qz_m": qz_m,
            "qz_v": qz_v,
            "z_expr": z_expr,
            "qzm_expr": qzm_expr,
            "qzv_expr": qzv_expr,
            "z_acc": z_acc,
            "qzm_acc": qzm_acc,
            "qzv_acc": qzv_acc,
            "z_pro": z_pro,
            "qzm_pro": qzm_pro,
            "qzv_pro": qzv_pro,
            "libsize_expr": libsize_expr,
            "libsize_acc": libsize_acc,
        }
        return outputs

    def _get_generative_input(self, tensors, inference_outputs, transform_batch=None):
        """Get the input for the generative model."""
        z = inference_outputs["z"]
        qz_m = inference_outputs["qz_m"]
        libsize_expr = inference_outputs["libsize_expr"]

        batch_index = tensors[REGISTRY_KEYS.BATCH_KEY]
        cont_key = REGISTRY_KEYS.CONT_COVS_KEY
        cont_covs = tensors[cont_key] if cont_key in tensors.keys() else None

        cat_key = REGISTRY_KEYS.CAT_COVS_KEY
        cat_covs = tensors[cat_key] if cat_key in tensors.keys() else None

        if transform_batch is not None:
            batch_index = torch.ones_like(batch_index) * transform_batch

        label = tensors[REGISTRY_KEYS.LABELS_KEY]

        input_dict = {
            "z": z,
            "qz_m": qz_m,
            "batch_index": batch_index,
            "cont_covs": cont_covs,
            "cat_covs": cat_covs,
            "libsize_expr": libsize_expr,
            "label": label,
        }
        return input_dict

    @auto_move_data
    def generative(
        self,
        z,
        qz_m,
        batch_index,
        cont_covs=None,
        cat_covs=None,
        libsize_expr=None,
        use_z_mean=False,
        label: torch.Tensor = None,
    ):
        """Runs the generative model."""
        if cat_covs is not None:
            categorical_input = torch.split(cat_covs, 1, dim=1)
        else:
            categorical_input = ()

        latent = z if not use_z_mean else qz_m
        if cont_covs is None:
            decoder_input = latent
        elif latent.dim() != cont_covs.dim():
            decoder_input = torch.cat(
                [latent, cont_covs.unsqueeze(0).expand(latent.size(0), -1, -1)], dim=-1
            )
        else:
            decoder_input = torch.cat([latent, cont_covs], dim=-1)

        # Accessibility Decoder
        p = self.z_decoder_accessibility(decoder_input, batch_index, *categorical_input)

        # Expression Decoder
        px_scale, _, px_rate, px_dropout = self.z_decoder_expression(
            self.gene_dispersion,
            decoder_input,
            libsize_expr,
            batch_index,
            *categorical_input,
            label,
        )
        # Expression Dispersion
        if self.gene_dispersion == "gene-label":
            px_r = F.linear(
                F.one_hot(label.squeeze(-1), self.n_labels).float(), self.px_r
            )  # px_r gets transposed - last dimension is nb genes
        elif self.gene_dispersion == "gene-batch":
            px_r = F.linear(F.one_hot(batch_index.squeeze(-1), self.n_batch).float(), self.px_r)
        elif self.gene_dispersion == "gene":
            px_r = self.px_r
        px_r = torch.exp(px_r)

        # Protein Decoder
        py_, log_pro_back_mean = self.z_decoder_pro(decoder_input, batch_index, *categorical_input)
        # Protein Dispersion
        if self.protein_dispersion == "protein-label":
            # py_r gets transposed - last dimension is n_proteins
            py_r = F.linear(F.one_hot(label.squeeze(-1), self.n_labels).float(), self.py_r)
        elif self.protein_dispersion == "protein-batch":
            py_r = F.linear(F.one_hot(batch_index.squeeze(-1), self.n_batch).float(), self.py_r)
        elif self.protein_dispersion == "protein":
            py_r = self.py_r
        py_r = torch.exp(py_r)
        py_["r"] = py_r

        return {
            "p": p,
            "px_scale": px_scale,
            "px_r": torch.exp(self.px_r),
            "px_rate": px_rate,
            "px_dropout": px_dropout,
            "py_": py_,
            "log_pro_back_mean": log_pro_back_mean,
        }

    def loss(self, tensors, inference_outputs, generative_outputs, kl_weight: float = 1.0):
        """Computes the loss function for the model."""
        # Get the data
        x = inference_outputs["x"]

        x_rna = x[:, : self.n_input_genes]
        x_atac = x[:, self.n_input_genes : (self.n_input_genes + self.n_input_regions)]
        if self.n_input_proteins == 0:
            y = torch.zeros(x.shape[0], 1, device=x.device, requires_grad=False)
        else:
            y = tensors[REGISTRY_KEYS.PROTEIN_EXP_KEY]

        mask_expr = x_rna.sum(dim=1) > 0
        mask_acc = x_atac.sum(dim=1) > 0
        mask_pro = y.sum(dim=1) > 0

        # Compute Accessibility loss
        p = generative_outputs["p"]
        libsize_acc = inference_outputs["libsize_acc"]
        rl_accessibility = self.get_reconstruction_loss_accessibility(x_atac, p, libsize_acc)

        # Compute Expression loss
        px_rate = generative_outputs["px_rate"]
        px_r = generative_outputs["px_r"]
        px_dropout = generative_outputs["px_dropout"]
        x_expression = x[:, : self.n_input_genes]
        rl_expression = self.get_reconstruction_loss_expression(
            x_expression, px_rate, px_r, px_dropout
        )

        # Compute Protein loss - No ability to mask minibatch (Param:None)
        if mask_pro.sum().gt(0):
            py_ = generative_outputs["py_"]
            rl_protein = get_reconstruction_loss_protein(y, py_, None)
        else:
            rl_protein = torch.zeros(x.shape[0], device=x.device, requires_grad=False)

        # calling without weights makes this act like a masked sum
        recon_loss_expression = rl_expression * mask_expr
        recon_loss_accessibility = rl_accessibility * mask_acc
        recon_loss_protein = rl_protein * mask_pro
        recon_loss = recon_loss_expression + recon_loss_accessibility + recon_loss_protein

        # Compute KLD between Z and N(0,I)
        qz_m = inference_outputs["qz_m"]
        qz_v = inference_outputs["qz_v"]
        kl_div_z = kld(
            Normal(qz_m, torch.sqrt(qz_v)),
            Normal(0, 1),
        ).sum(dim=1)

        # Compute KLD between distributions for paired data
        kl_div_paired = self._compute_mod_penalty(
            (inference_outputs["qzm_expr"], inference_outputs["qzv_expr"]),
            (inference_outputs["qzm_acc"], inference_outputs["qzv_acc"]),
            (inference_outputs["qzm_pro"], inference_outputs["qzv_pro"]),
            mask_expr,
            mask_acc,
            mask_pro,
        )

        # KL WARMUP
        kl_local_for_warmup = kl_div_z
        weighted_kl_local = kl_weight * kl_local_for_warmup + kl_div_paired

        # TOTAL LOSS
        loss = torch.mean(recon_loss + weighted_kl_local)

        recon_losses = {
            "reconstruction_loss_expression": recon_loss_expression,
            "reconstruction_loss_accessibility": recon_loss_accessibility,
            "reconstruction_loss_protein": recon_loss_protein,
        }
        kl_local = {
            "kl_divergence_z": kl_div_z,
            "kl_divergence_paired": kl_div_paired,
        }
        return LossOutput(loss=loss, reconstruction_loss=recon_losses, kl_local=kl_local)

    def get_reconstruction_loss_expression(self, x, px_rate, px_r, px_dropout):
        """Computes the reconstruction loss for the expression data."""
        rl = 0.0
        if self.gene_likelihood == "zinb":
            rl = (
                -ZeroInflatedNegativeBinomial(mu=px_rate, theta=px_r, zi_logits=px_dropout)
                .log_prob(x)
                .sum(dim=-1)
            )
        elif self.gene_likelihood == "nb":
            rl = -NegativeBinomial(mu=px_rate, theta=px_r).log_prob(x).sum(dim=-1)
        elif self.gene_likelihood == "poisson":
            rl = -Poisson(px_rate).log_prob(x).sum(dim=-1)
        return rl

    def get_reconstruction_loss_accessibility(self, x, p, d):
        """Computes the reconstruction loss for the accessibility data."""
        reg_factor = torch.sigmoid(self.region_factors) if self.region_factors is not None else 1
        return torch.nn.BCELoss(reduction="none")(p * d * reg_factor, (x > 0).float()).sum(dim=-1)

    def _compute_mod_penalty(self, mod_params1, mod_params2, mod_params3, mask1, mask2, mask3):
        """Computes Similarity Penalty across modalities given selection (None, Jeffreys, MMD).

        Parameters
        ----------
        mod_params1/2/3
            Posterior parameters for for modality 1/2/3
        mask1/2/3
            mask for modality 1/2/3
        """
        mask12 = torch.logical_and(mask1, mask2)
        mask13 = torch.logical_and(mask1, mask3)
        mask23 = torch.logical_and(mask3, mask2)

        if self.modality_penalty == "None":
            return 0
        elif self.modality_penalty == "Jeffreys":
            pair_penalty = torch.zeros(mask1.shape[0], device=mask1.device, requires_grad=True)
            if mask12.sum().gt(0):
                penalty12 = sym_kld(
                    mod_params1[0],
                    mod_params1[1].sqrt(),
                    mod_params2[0],
                    mod_params2[1].sqrt(),
                )
                penalty12 = torch.where(mask12, penalty12.T, torch.zeros_like(penalty12).T).sum(
                    dim=0
                )
                pair_penalty = pair_penalty + penalty12
            if mask13.sum().gt(0):
                penalty13 = sym_kld(
                    mod_params1[0],
                    mod_params1[1].sqrt(),
                    mod_params3[0],
                    mod_params3[1].sqrt(),
                )
                penalty13 = torch.where(mask13, penalty13.T, torch.zeros_like(penalty13).T).sum(
                    dim=0
                )
                pair_penalty = pair_penalty + penalty13
            if mask23.sum().gt(0):
                penalty23 = sym_kld(
                    mod_params2[0],
                    mod_params2[1].sqrt(),
                    mod_params3[0],
                    mod_params3[1].sqrt(),
                )
                penalty23 = torch.where(mask23, penalty23.T, torch.zeros_like(penalty23).T).sum(
                    dim=0
                )
                pair_penalty = pair_penalty + penalty23

        elif self.modality_penalty == "MMD":
            pair_penalty = torch.zeros(mask1.shape[0], device=mask1.device, requires_grad=True)
            if mask12.sum().gt(0):
                penalty12 = torch.linalg.norm(mod_params1[0] - mod_params2[0], dim=1)
                penalty12 = torch.where(mask12, penalty12.T, torch.zeros_like(penalty12).T).sum(
                    dim=0
                )
                pair_penalty = pair_penalty + penalty12
            if mask13.sum().gt(0):
                penalty13 = torch.linalg.norm(mod_params1[0] - mod_params3[0], dim=1)
                penalty13 = torch.where(mask13, penalty13.T, torch.zeros_like(penalty13).T).sum(
                    dim=0
                )
                pair_penalty = pair_penalty + penalty13
            if mask23.sum().gt(0):
                penalty23 = torch.linalg.norm(mod_params2[0] - mod_params3[0], dim=1)
                penalty23 = torch.where(mask23, penalty23.T, torch.zeros_like(penalty23).T).sum(
                    dim=0
                )
                pair_penalty = pair_penalty + penalty23
        else:
            raise ValueError("modality penalty not supported")

        return pair_penalty


@auto_move_data
def mix_modalities(Xs, masks, weights, weight_transform: callable = None):
    """Compute the weighted mean of the Xs while masking unmeasured modality values.

    Parameters
    ----------
    Xs
        Sequence of Xs to mix, each should be (N x D)
    masks
        Sequence of masks corresponding to the Xs, indicating whether the values
        should be included in the mix or not (N)
    weights
        Weights for each modality (either K or N x K)
    weight_transform
        Transformation to apply to the weights before using them
    """
    # (batch_size x latent) -> (batch_size x modalities x latent)
    Xs = torch.stack(Xs, dim=1)
    # (batch_size) -> (batch_size x modalities)
    masks = torch.stack(masks, dim=1).float()
    weights = masked_softmax(weights, masks, dim=-1)

    # (batch_size x modalities) -> (batch_size x modalities x latent)
    weights = weights.unsqueeze(-1)
    if weight_transform is not None:
        weights = weight_transform(weights)

    # sum over modalities, so output is (batch_size x latent)
    return (weights * Xs).sum(1)


@auto_move_data
def sym_kld(qzm1, qzv1, qzm2, qzv2):
    """Symmetric KL divergence between two Gaussians."""
    rv1 = Normal(qzm1, qzv1.sqrt())
    rv2 = Normal(qzm2, qzv2.sqrt())

    return kld(rv1, rv2) + kld(rv2, rv1)


@auto_move_data
def get_reconstruction_loss_protein(y, py_, pro_batch_mask_minibatch=None):
    """Get the reconstruction loss for protein data."""
    py_conditional = NegativeBinomialMixture(
        mu1=py_["rate_back"],
        mu2=py_["rate_fore"],
        theta1=py_["r"],
        mixture_logits=py_["mixing"],
    )

    reconst_loss_protein_full = -py_conditional.log_prob(y)

    if pro_batch_mask_minibatch is not None:
        temp_pro_loss_full = pro_batch_mask_minibatch.bool() * reconst_loss_protein_full
        rl_protein = temp_pro_loss_full.sum(dim=-1)
    else:
        rl_protein = reconst_loss_protein_full.sum(dim=-1)

    return rl_protein



import json
import os

import pandas as pd
from mudata import MuData
from scipy.stats import pearsonr, spearmanr
from sklearn.metrics import mean_absolute_error as mae
from sklearn.metrics import r2_score

from scvi._types import AnnOrMuData
from scvi.model._utils import REGISTRY_KEYS
from scvi.model.base import BaseModelClass

from ._ppc import PosteriorPredictiveCheck as PPC

METRIC_CV_CELL = "cv_cell"
METRIC_CV_GENE = "cv_gene"
METRIC_DIFF_EXP = "diff_exp"


def _dataframe_to_markdown(df):
    # Create the header
    header = "| Index | " + " | ".join(df.columns) + " |"
    separator = "| --- | " + " | ".join("---" for _ in df.columns) + " |"

    # Format values and create rows, including the index
    rows = "\n".join(
        "| "
        + " | ".join(
            [str(index)]
            + [f"{value:.2f}" if not isinstance(value, int) else f"{value}" for value in row]
        )
        + " |"
        for index, row in zip(df.index, df.values, strict=True)
    )

    # Combine header, separator, and rows
    return f"{header}\n{separator}\n{rows}"


def create_criticism_report(
    model: BaseModelClass,
    adata: AnnOrMuData | None = None,
    skip_metrics: list = (),
    n_samples: int = 5,
    label_key: str | None = None,
    save_folder: str | None = None,
) -> dict:
    """
    Helper function to compute and store criticism metrics for a model.

    Parameters
    ----------
    model
        Model to compute metrics on.
    adata
        AnnOrMuData to compute metrics on.
    skip_metrics
        List of metrics to skip. Can be one of ["cv_cell", "cv_gene", "diff_exp"].
    n_samples
        Number of samples to use for posterior predictive check.
    label_key
        Key in adata.obs to use as cell type labels. If None, will use the original label key from
        the model.
    save_folder
        Path to folder for storing the metrics. Preferred to store in save_path folder of model.
    """
    adata = model._validate_anndata(adata)

    if isinstance(adata, MuData):
        modalities = model.registry_["setup_args"]["modalities"]
        modalities = [modalities[i] for i in modalities if "layer" in i]
        md_cell_wise_cv, md_gene_wise_cv, md_de = "", "", ""
        for i in modalities:
            md_cell_wise_cv_, md_gene_wise_cv_, md_de_ = compute_metrics(
                model, adata, skip_metrics, n_samples, label_key, modality=i
            )
            md_cell_wise_cv += f"Modality: {i}\n\n" + md_cell_wise_cv_ + "\n\n"
            md_gene_wise_cv += f"Modality: {i}\n\n" + md_gene_wise_cv_ + "\n\n"
            md_de += f"Modality: {i}\n\n" + md_de_ + "\n\n"
    else:
        md_cell_wise_cv, md_gene_wise_cv, md_de = compute_metrics(
            model, adata, skip_metrics, n_samples, label_key
        )

    markdown_dict = {
        "cell_wise_cv": md_cell_wise_cv,
        "gene_wise_cv": md_gene_wise_cv,
        "diff_exp": md_de,
    }

    save_path = os.path.join(save_folder, "metrics.json")

    with open(save_path, "w") as f:
        json.dump(markdown_dict, f, indent=4)


def compute_metrics(model, adata, skip_metrics, n_samples, label_key, modality=None):
    models_dict = {"model": model}
    ppc = PPC(adata, models_dict, n_samples=n_samples, modality=modality)
    # run ppc+cv
    if METRIC_CV_CELL not in skip_metrics:
        ppc.coefficient_of_variation("features")
        md_cell_wise_cv = _cv_metrics(ppc, model, cell_wise=True)
    else:
        md_cell_wise_cv = ""
    if METRIC_CV_GENE not in skip_metrics:
        ppc.coefficient_of_variation("cells")
        md_gene_wise_cv = _cv_metrics(ppc, model, cell_wise=False)
    else:
        md_gene_wise_cv = ""
    # run diff_exp
    if METRIC_DIFF_EXP not in skip_metrics:
        labels_state_registry = model.adata_manager.get_state_registry(REGISTRY_KEYS.LABELS_KEY)
        if label_key is None and labels_state_registry.original_key != "_scvi_labels":
            label_key = labels_state_registry.original_key
        ppc.differential_expression(de_groupby=label_key, p_val_thresh=0.2)
        summary_df = ppc.metrics[METRIC_DIFF_EXP]["summary"].set_index("group")
        summary_df = summary_df.drop(columns=["model"])
        summary_df = summary_df.sort_values(by="n_cells", ascending=False)
        md_de = _dataframe_to_markdown(summary_df)
    if modality is not None:
        adata = model.adata[modality]
    else:
        adata = model.adata
    adata.var["cv_gene_ratio"] = ppc.metrics[METRIC_CV_GENE]["model"] / (
        ppc.metrics[METRIC_CV_GENE]["model"] + ppc.metrics[METRIC_CV_GENE]["Raw"]
    )
    adata.obs["cv_cell_ratio"] = ppc.metrics[METRIC_CV_CELL]["model"] / (
        ppc.metrics[METRIC_CV_CELL]["model"] + ppc.metrics[METRIC_CV_CELL]["Raw"]
    )
    adata.varm["lfc_model"] = pd.DataFrame(
        {
            key: value["approx"]
            for key, value in ppc.metrics["diff_exp"]["lfc_per_model_per_group"]["model"].items()
        }
    ).loc[adata.var_names]
    adata.varm["lfc_raw"] = pd.DataFrame(
        {
            key: value["raw"]
            for key, value in ppc.metrics["diff_exp"]["lfc_per_model_per_group"]["model"].items()
        }
    ).loc[adata.var_names]

    return md_cell_wise_cv, md_gene_wise_cv, md_de


def _cv_metrics(ppc, model, cell_wise: bool = True):
    """
    Compute metrics of coefficient of variation.

    Parameters
    ----------
    model_name
        Name of the model
    cell_wise
        Whether to plot the cell-wise or gene-wise metric
    plt_type
        The type of plot to generate.
    model
        Model used for generating metric to split validation and training data
    indices
        Indices of the data to plot, if None, all data will be plotted
    """
    metric = METRIC_CV_CELL if cell_wise is True else METRIC_CV_GENE
    model_metric = ppc.metrics[metric]["model"].values
    raw_metric = ppc.metrics[metric]["Raw"].values

    if hasattr(model, "train_indices"):
        train_indices = model.train_indices
        validation_indices = model.validation_indices
        if train_indices is None:
            train_indices = []
        if validation_indices is None:
            validation_indices = []
    else:
        train_indices = []
        validation_indices = []

    if cell_wise and len(train_indices) > 0 and len(validation_indices) > 0:
        indices = [model.train_indices, model.validation_indices]

        mae_values = [
            mae(model_metric[indices[0]], raw_metric[indices[0]]),
            mae(model_metric[indices[1]], raw_metric[indices[1]]),
        ]
        pearsonr_values = [
            pearsonr(model_metric[indices[0]], raw_metric[indices[0]])[0],
            pearsonr(model_metric[indices[1]], raw_metric[indices[1]])[0],
        ]
        spearmanr_values = [
            spearmanr(model_metric[indices[0]], raw_metric[indices[0]])[0],
            spearmanr(model_metric[indices[1]], raw_metric[indices[1]])[0],
        ]
        r2_score_values = [
            r2_score(model_metric[indices[0]], raw_metric[indices[0]]),
            r2_score(model_metric[indices[1]], raw_metric[indices[1]]),
        ]
        metric_table = (
            "| Metric                  | Training Value | Validation Value |\n"
            "|-------------------------|----------------|------------------|\n"
            f"| Mean Absolute Error | {mae_values[0]:.2f}  | {mae_values[1]:.2f}           |\n"
            f"| Pearson Correlation | {pearsonr_values[0]:.2f}  | {pearsonr_values[1]:.2f}  |\n"
            f"| Spearman Correlation | {spearmanr_values[0]:.2f} | {spearmanr_values[1]:.2f}  |\n"
            f"| R² (R-Squared) | {r2_score_values[0]:.2f}  | {r2_score_values[1]:.2f}      |"
        )
    else:
        indices = [model.train_indices, model.validation_indices]

        mae_values = mae(model_metric, raw_metric)
        pearsonr_values = pearsonr(model_metric, raw_metric)[0]
        spearmanr_values = spearmanr(model_metric, raw_metric)[0]
        r2_score_values = r2_score(model_metric, raw_metric)
        metric_table = (
            "| Metric                  | Training Value |\n"
            "|-------------------------|----------------|\n"
            f"| Mean Absolute Error | {mae_values:.2f}   |\n"
            f"| Pearson Correlation | {pearsonr_values:.2f}  |\n"
            f"| Spearman Correlation | {spearmanr_values:.2f} |\n"
            f"| R² (R-Squared) | {r2_score_values:.2f}  |"
        )

    return metric_table



from collections.abc import Callable, Mapping, Sequence
from functools import wraps
from typing import Any

import flax.linen as nn
import torch
from torch.nn import Module


def auto_move_data(fn: Callable) -> Callable:
    """Decorator for :class:`~torch.nn.Module` methods to move data to correct device.

    Input arguments are moved automatically to the correct device.
    It has no effect if applied to a method of an object that is not an instance of
    :class:`~torch.nn.Module` and is typically applied to ``__call__``
    or ``forward``.

    Parameters
    ----------
    fn
        A nn.Module method for which the arguments should be moved to the device
        the parameters are on.
    """

    @wraps(fn)
    def auto_transfer_args(self, *args, **kwargs):
        if not isinstance(self, Module):
            return fn(self, *args, **kwargs)

        # decorator only necessary after training
        if self.training:
            return fn(self, *args, **kwargs)

        device = list({p.device for p in self.parameters()})
        if len(device) > 1:
            raise RuntimeError("Module tensors on multiple devices.")
        else:
            device = device[0]
        args = _move_data_to_device(args, device)
        kwargs = _move_data_to_device(kwargs, device)
        return fn(self, *args, **kwargs)

    return auto_transfer_args


def _move_data_to_device(batch: Any, device: torch.device):
    """Transfers a collection of data to the given device.

    Any object that defines a method ``to(device)`` will be moved and all other objects
    in the collection will be left untouched.

    Parameters
    ----------
    batch
        A tensor or collection of tensors or anything that has a method ``.to(...)``.
        See :meth:`apply_to_collection` for a list of supported collection types.
    device
        The device to which the data should be moved

    Returns
    -------
        The same collection but with all contained tensors residing on the new device.
    """

    def batch_to(data):
        kwargs = {"non_blocking": True} if isinstance(data, torch.Tensor) else {}
        return data.to(device, **kwargs)

    return _apply_to_collection(batch, dtype=torch.Tensor, function=batch_to)


def _apply_to_collection(
    data: Any,
    dtype: type | tuple,
    function: Callable,
    *args,
    **kwargs,
) -> Any:
    """Recursively applies a function to all elements of a certain dtype.

    Parameters
    ----------
    data
        The collection to apply the function to.
    dtype
        The given function will be applied to all elements of this dtype.
    function
        The function to apply.
    *args
        Positional arguments to be passed to the function.
    **kwargs
        Keyword arguments to be passed to the function.

    Returns
    -------
    The resulting collection.
    """
    elem_type = type(data)

    # Breaking condition
    if isinstance(data, dtype):
        return function(data, *args, **kwargs)

    # Recursively apply to collection items
    elif isinstance(data, Mapping):
        return elem_type(
            {k: _apply_to_collection(v, dtype, function, *args, **kwargs) for k, v in data.items()}
        )
    elif isinstance(data, tuple) and hasattr(data, "_fields"):  # named tuple
        return elem_type(
            *(_apply_to_collection(d, dtype, function, *args, **kwargs) for d in data)
        )
    elif isinstance(data, Sequence) and not isinstance(data, str):
        return elem_type([_apply_to_collection(d, dtype, function, *args, **kwargs) for d in data])

    # data is neither of dtype, nor a collection
    return data


def flax_configure(cls: nn.Module) -> Callable:
    """Decorator to raise an error if a boolean `training` param is missing in the call."""
    original_init = cls.__init__

    @wraps(original_init)
    def init(self, *args, **kwargs):
        self.configure()
        original_init(self, *args, **kwargs)
        if not isinstance(self.training, bool):
            raise ValueError("Custom sublclasses must have a training parameter.")

    cls.__init__ = init
    return cls



import pytest
from anndata import AnnData

from scvi import REGISTRY_KEYS
from scvi.data import AnnDataManager, fields, synthetic_iid
from scvi.data._utils import _get_adata_minify_type
from scvi.model.base import BaseModelClass


class TestModelClass(BaseModelClass):
    @classmethod
    def setup_anndata(
        cls,
        adata: AnnData,
        layer: str | None = None,
        batch_key: str | None = None,
        labels_key: str | None = None,
        size_factor_key: str | None = None,
        categorical_covariate_keys: list[str] | None = None,
        continuous_covariate_keys: list[str] | None = None,
        **kwargs,
    ):
        setup_method_args = cls._get_setup_method_args(**locals())
        anndata_fields = [
            fields.LayerField(REGISTRY_KEYS.X_KEY, layer, is_count_data=True),
            fields.CategoricalObsField(REGISTRY_KEYS.BATCH_KEY, batch_key),
            fields.CategoricalObsField(REGISTRY_KEYS.LABELS_KEY, labels_key),
            fields.NumericalObsField(
                REGISTRY_KEYS.SIZE_FACTOR_KEY, size_factor_key, required=False
            ),
            fields.CategoricalJointObsField(
                REGISTRY_KEYS.CAT_COVS_KEY, categorical_covariate_keys
            ),
            fields.NumericalJointObsField(REGISTRY_KEYS.CONT_COVS_KEY, continuous_covariate_keys),
        ]
        # register new fields if the adata is minified
        adata_minify_type = _get_adata_minify_type(adata)
        if adata_minify_type is not None:
            anndata_fields += cls._get_fields_for_adata_minification(adata_minify_type)
        adata_manager = AnnDataManager(fields=anndata_fields, setup_method_args=setup_method_args)
        adata_manager.register_fields(adata, **kwargs)
        cls.register_manager(adata_manager)

    def train(self):
        pass


def test_deregister_manager():
    adata = synthetic_iid()
    bdata = synthetic_iid()

    # default deregister
    TestModelClass.setup_anndata(adata)
    TestModelClass.setup_anndata(bdata)
    model = TestModelClass(adata)
    adata_manager = model._get_most_recent_anndata_manager(adata)
    bdata_manager = model._get_most_recent_anndata_manager(bdata)

    model.deregister_manager()
    instance_manager_store = TestModelClass._per_instance_manager_store[model.id]
    class_manager_store = model._setup_adata_manager_store
    assert adata_manager.adata_uuid in instance_manager_store
    assert adata_manager.adata_uuid in class_manager_store
    assert bdata_manager.adata_uuid not in class_manager_store

    with pytest.raises(ValueError):
        model.deregister_manager(bdata)

    # deregister with argument
    TestModelClass.setup_anndata(adata)
    TestModelClass.setup_anndata(bdata)
    model = TestModelClass(adata)
    adata_manager = model._get_most_recent_anndata_manager(adata)
    bdata_manager = model._get_most_recent_anndata_manager(bdata)

    model.deregister_manager(adata)
    instance_manager_store = TestModelClass._per_instance_manager_store[model.id]
    class_manager_store = model._setup_adata_manager_store
    assert adata_manager.adata_uuid not in instance_manager_store
    assert adata_manager.adata_uuid not in class_manager_store
    assert bdata_manager.adata_uuid in class_manager_store
    model.deregister_manager(bdata)
    assert bdata_manager.adata_uuid not in class_manager_store

    with pytest.raises(ValueError):
        model.deregister_manager(adata)
    with pytest.raises(ValueError):
        model.deregister_manager(bdata)



import warnings
from itertools import cycle

from scvi import settings
from scvi.data import AnnDataManager
from scvi.dataloaders import ConcatDataLoader


class _ContrastiveIterator:
    """Iterator for background and target dataloader pair in contrastive analysis.

    Each iteration of this iterator returns a dictionary with two elements:
    "background", containing one batch of data from the background dataloader, and
    "target", containing one batch of data from the target dataloader.
    """

    def __init__(self, background, target):
        self.background = iter(background)
        self.target = iter(target)

    def __iter__(self):
        return self

    def __next__(self):
        bg_samples = next(self.background)
        tg_samples = next(self.target)
        return {"background": bg_samples, "target": tg_samples}


class ContrastiveDataLoader(ConcatDataLoader):
    """Dataloader to load background and target data for contrastive analysis.

    Each iteration of the dataloader returns a dictionary containing background and
    target data points, indexed by "background" and "target", respectively.

    Parameters
    ----------
        adata_manager
            :class:`~scvi.data.AnnDataManager` object that has been created via
            ``setup_anndata``.
        background_indices
            Indices for background samples in the adata.
        target_indices
            Indices for target samples in the adata.
        shuffle
            Whether the data should be shuffled.
        batch_size
            Mini-batch size to load for background and target data.
        data_and_attributes
            Dictionary with keys representing keys in data registry
            (``adata_manager.data_registry``) and value equal to desired numpy loading
            type (later made into torch tensor). If ``None``, defaults to all registered
            data.
        drop_last
            If ``int``, drops the last batch if its length is less than
            ``drop_last``. If ``drop_last == True``, drops last non-full batch.
            If ``drop_last == False``, iterate over all batches.
        distributed_sampler
            ``EXPERIMENTAL`` Whether to use :class:`~scvi.dataloaders.BatchDistributedSampler` as
            the sampler. If `True`, `sampler` must be `None`. Not applicable here.
        load_sparse_tensor
            ``EXPERIMENTAL`` If ``True``, loads data with sparse CSR or CSC layout as a
            :class:`~torch.Tensor` with the same layout. Can lead to speedups in data
            transfers to GPUs, depending on the sparsity of the data. Not applicable
            here.
        **data_loader_kwargs: Keyword arguments for `torch.utils.data.DataLoader`.
    """

    def __init__(
        self,
        adata_manager: AnnDataManager,
        background_indices: list[int],
        target_indices: list[int],
        shuffle: bool = False,
        batch_size: int = 128,
        data_and_attributes: dict | None = None,
        drop_last: bool | int = False,
        distributed_sampler: bool = False,
        load_sparse_tensor: bool = False,
        **data_loader_kwargs,
    ) -> None:
        super().__init__(
            adata_manager=adata_manager,
            indices_list=[background_indices, target_indices],
            shuffle=shuffle,
            batch_size=batch_size,
            data_and_attributes=data_and_attributes,
            drop_last=drop_last,
            **data_loader_kwargs,
        )
        self.background_indices = background_indices
        self.target_indices = target_indices
        if distributed_sampler:
            warnings.warn(
                (
                    "distributed_sampler=True is not implemented for "
                    "ContrastiveDataLoader. Setting distributed_sampler=False"
                ),
                UserWarning,
                stacklevel=settings.warnings_stacklevel,
            )
            distributed_sampler = False
        if load_sparse_tensor:
            warnings.warn(
                (
                    "load_sparse_tensor=True is not implemented for "
                    "ContrastiveDataLoader. Setting load_sparse_tensor=False"
                ),
                UserWarning,
                stacklevel=settings.warnings_stacklevel,
            )
            load_sparse_tensor = False
        self.distributed_sampler = distributed_sampler
        self.load_sparse_tensor = load_sparse_tensor

    def __iter__(self):
        """Iter method for contrastive dataloader.

        Will iter over the dataloader with the most data while cycling through
        the data in the other dataloader.
        """
        iter_list = [cycle(dl) if dl != self.largest_dl else dl for dl in self.dataloaders]
        return _ContrastiveIterator(background=iter_list[0], target=iter_list[1])



import pytest
import torch

from scvi.nn import one_hot


def test_one_hot_deprecation_warning():
    """Public API should give deprecation warning"""
    x = torch.tensor([[1], [1], [2]])
    with pytest.warns(DeprecationWarning):
        one_hot(x, 3)



from collections.abc import Sequence

import numpy as np
import pyro
import torch
import torch.nn.functional as F
from anndata import AnnData

from scvi._constants import REGISTRY_KEYS
from scvi.data import AnnDataManager
from scvi.data.fields import LayerField
from scvi.model.base import BaseModelClass, PyroSviTrainMixin
from scvi.train import PyroTrainingPlan
from scvi.utils import setup_anndata_dsp

from ._module import DecipherPyroModule
from ._trainingplan import DecipherTrainingPlan
from .utils._trajectory import Trajectory


class Decipher(PyroSviTrainMixin, BaseModelClass):
    """Decipher model for single-cell data analysis :cite:p:`Nazaret23`.

    Parameters
    ----------
    adata
        AnnData object that has been registered via
        :meth:`~scvi.model.Decipher.setup_anndata`.
    dim_v
        Dimension of the interpretable latent space v.
    dim_z
        Dimension of the intermediate latent space z.
    layers_v_to_z
        Hidden layer sizes for the v to z decoder network.
    layers_z_to_x
        Hidden layer sizes for the z to x decoder network.
    beta
        Regularization parameter for the KL divergence.
    """

    _module_cls = DecipherPyroModule
    _training_plan_cls = DecipherTrainingPlan

    def __init__(self, adata: AnnData, **kwargs):
        pyro.clear_param_store()

        super().__init__(adata)

        dim_genes = self.summary_stats.n_vars

        self.module = self._module_cls(
            dim_genes,
            **kwargs,
        )

        self.init_params_ = self._get_init_params(locals())

    @classmethod
    @setup_anndata_dsp.dedent
    def setup_anndata(
        cls,
        adata: AnnData,
        layer: str | None = None,
        **kwargs,
    ) -> AnnData | None:
        """%(summary)s.

        Parameters
        ----------
        %(param_adata)s
        %(param_layer)s
        """
        setup_method_args = cls._get_setup_method_args(**locals())
        anndata_fields = [
            LayerField(REGISTRY_KEYS.X_KEY, layer, is_count_data=True),
        ]
        adata_manager = AnnDataManager(fields=anndata_fields, setup_method_args=setup_method_args)
        adata_manager.register_fields(adata, **kwargs)
        cls.register_manager(adata_manager)

    def train(
        self,
        max_epochs: int | None = None,
        accelerator: str = "auto",
        device: int | str = "auto",
        train_size: float = 0.9,
        validation_size: float | None = None,
        shuffle_set_split: bool = True,
        batch_size: int = 128,
        early_stopping: bool = False,
        training_plan: PyroTrainingPlan | None = None,
        datasplitter_kwargs: dict | None = None,
        plan_kwargs: dict | None = None,
        **trainer_kwargs,
    ):
        if "early_stopping_monitor" not in trainer_kwargs:
            trainer_kwargs["early_stopping_monitor"] = "nll_validation"
        datasplitter_kwargs = datasplitter_kwargs or {}
        if "drop_last" not in datasplitter_kwargs:
            datasplitter_kwargs["drop_last"] = True
        super().train(
            max_epochs=max_epochs,
            accelerator=accelerator,
            device=device,
            train_size=train_size,
            validation_size=validation_size,
            shuffle_set_split=shuffle_set_split,
            batch_size=batch_size,
            early_stopping=early_stopping,
            plan_kwargs=plan_kwargs,
            training_plan=training_plan,
            datasplitter_kwargs=datasplitter_kwargs,
            **trainer_kwargs,
        )

    def get_latent_representation(
        self,
        adata: AnnData | None = None,
        indices: Sequence[int] | None = None,
        batch_size: int | None = None,
        give_z: bool = False,
    ) -> np.ndarray:
        """Get the latent representation of the data.

        Parameters
        ----------
        adata
            AnnData object with the data to get the latent representation of.
        indices
            Indices of the data to get the latent representation of.
        batch_size
            Batch size to use for the data loader.
        give_z
            Whether to return the intermediate latent space z or the top-level
            latent space v.
        """
        self._check_if_trained(warn=False)
        adata = self._validate_anndata(adata)

        scdl = self._make_data_loader(adata=adata, indices=indices, batch_size=batch_size)
        latent_locs = []
        for tensors in scdl:
            x = tensors[REGISTRY_KEYS.X_KEY]
            x = torch.log1p(x)
            x = x.to(self.module.device)
            z_loc, _ = self.module.encoder_x_to_z(x)
            if give_z:
                latent_locs.append(z_loc)
            else:
                v_loc, _ = self.module.encoder_zx_to_v(torch.cat([z_loc, x], dim=-1))
                latent_locs.append(v_loc)
        return torch.cat(latent_locs).detach().cpu().numpy()

    def compute_imputed_gene_expression(
        self,
        adata: AnnData | None = None,
        indices: Sequence[int] | None = None,
        batch_size: int | None = None,
        compute_covariances: bool = False,
        v_obsm_key: str | None = None,
        z_obsm_key: str | None = None,
    ) -> np.ndarray | tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Impute gene expression from the decipher model.

        Parameters
        ----------
        adata
            The annotated data matrix.
        indices
            Indices of the data to get the latent representation of.
        batch_size
            Batch size to use for the data loader.
        compute_covariances
            Whether to compute the covariances between the Decipher v and each gene.
        v_obsm_key
            Key in `adata.obsm` to use for the Decipher v. Required if
            `compute_covariances` is True.
        z_obsm_key
            Key in `adata.obsm` to use for the Decipher z. Required if
            `compute_covariances` is True.

        Returns
        -------
        The imputed gene expression, and the covariances between the Decipher v and each gene
        if `compute_covariances` is True.
        """
        if compute_covariances and (v_obsm_key is None or z_obsm_key is None):
            raise ValueError(
                "`v_obsm_key` and `z_obsm_key` must be provided if `compute_covariances` is True."
            )

        adata = self._validate_anndata(adata)
        scdl = self._make_data_loader(adata=adata, indices=indices, batch_size=batch_size)

        imputed_gene_expression_batches = []
        for tensors in scdl:
            x = tensors[REGISTRY_KEYS.X_KEY]
            z_loc, _, _, _ = self.module.guide(x)
            mu = self.module.decoder_z_to_x(z_loc)
            mu = F.softmax(mu, dim=-1)
            library_size = x.sum(axis=-1, keepdim=True)
            imputed_gene_expr = (library_size * mu.detach().cpu()).numpy()
            imputed_gene_expression_batches.append(imputed_gene_expr)
        imputed_gene_expression = np.concatenate(imputed_gene_expression_batches, axis=0)

        if compute_covariances:
            G = imputed_gene_expression.shape[1]
            v_gene_covariance = np.cov(
                imputed_gene_expression,
                y=adata.obsm[v_obsm_key],
                rowvar=False,
            )[:G, G:]
            z_gene_covariance = np.cov(
                imputed_gene_expression,
                y=adata.obsm[z_obsm_key],
                rowvar=False,
            )[:G, G:]
            return imputed_gene_expression, v_gene_covariance, z_gene_covariance

        return imputed_gene_expression

    @staticmethod
    def compute_decipher_time(
        adata: AnnData,
        cluster_obs_key: str,
        trajectory: Trajectory,
        n_neighbors: int = 10,
    ) -> np.ndarray:
        """Compute the decipher time for each cell, based on the inferred trajectories.

        The decipher time is computed by KNN regression of the cells'
        decipher v on the trajectories.

        Parameters
        ----------
        adata : AnnData
            The annotated data matrix.
        cluster_obs_key : str
            The key in adata.obs containing cluster assignments.
        trajectory : Trajectory
            A Trajectory object containing the trajectory information.
        n_neighbors : int
            The number of neighbors to use for the KNN regression.

        Returns
        -------
        The decipher time of each cell.
        """
        try:
            from sklearn.neighbors import KNeighborsRegressor
        except ImportError as err:
            raise ImportError("Please install scikit-learn -- `pip install scikit-learn`") from err

        decipher_time = np.full(adata.n_obs, np.nan)

        knn = KNeighborsRegressor(n_neighbors=n_neighbors)
        knn.fit(trajectory.trajectory_latent, trajectory.trajectory_time)
        is_on_trajectory = adata.obs[cluster_obs_key].isin(trajectory.cluster_ids)
        cells_on_trajectory_idx = np.where(is_on_trajectory)[0]

        decipher_time[cells_on_trajectory_idx] = knn.predict(
            adata.obsm[trajectory.rep_key][cells_on_trajectory_idx]
        )
        return decipher_time

    def compute_gene_patterns(
        self,
        adata: AnnData,
        trajectory: Trajectory,
        l_scale: float = 10_000,
        n_samples: int = 100,
    ) -> dict[str, np.ndarray]:
        """Compute the gene patterns for a trajectory.

        The trajectory's points are sent through the decoders, thus defining distributions over the
        gene expression. The gene patterns are computed by sampling from these distribution.

        Parameters
        ----------
        adata : AnnData
            The annotated data matrix.
        trajectory : Trajectory
            A Trajectory object containing the trajectory information.
        l_scale : float
            The library size scaling factor.
        n_samples : int
            The number of samples to draw from the decoder to compute the gene pattern statistics.

        Returns
        -------
        The gene patterns for the trajectory.
        Dictionary keys:
            - `mean`: the mean gene expression pattern
            - `q25`: the 25% quantile of the gene expression pattern
            - `q75`: the 75% quantile of the gene expression pattern
            - `times`: the times of the trajectory
        """
        adata = self._validate_anndata(adata)

        t_points = trajectory.trajectory_latent
        t_times = trajectory.trajectory_time

        t_points = torch.FloatTensor(t_points).to(self.module.device)
        z_mean, z_scale = self.module.decoder_v_to_z(t_points)
        z_scale = F.softplus(z_scale)

        z_samples = torch.distributions.Normal(z_mean, z_scale).sample(sample_shape=(n_samples,))

        gene_patterns = {}
        gene_patterns["mean"] = (
            F.softmax(self.module.decoder_z_to_x(z_mean), dim=-1).detach().cpu().numpy() * l_scale
        )

        gene_expression_samples = (
            F.softmax(self.module.decoder_z_to_x(z_samples), dim=-1).detach().cpu().numpy()
            * l_scale
        )
        gene_patterns["q25"] = np.quantile(gene_expression_samples, 0.25, axis=0)
        gene_patterns["q75"] = np.quantile(gene_expression_samples, 0.75, axis=0)

        gene_patterns["times"] = t_times

        return gene_patterns



from pyro.infer import Predictive

from ._decorators import auto_move_data


class AutoMoveDataPredictive(Predictive):
    """Auto move data."""

    @auto_move_data
    def forward(self, *args, **kwargs):
        return super().forward(*args, **kwargs)



from ._contrastive_data_splitting import ContrastiveDataSplitter
from ._contrastive_dataloader import ContrastiveDataLoader
from ._model import ContrastiveVI
from ._module import ContrastiveVAE

__all__ = [
    "ContrastiveDataLoader",
    "ContrastiveDataSplitter",
    "ContrastiveVAE",
    "ContrastiveVI",
]



from __future__ import annotations

import os
import re
from typing import TYPE_CHECKING

from anndata import read_h5ad

from scvi.utils import dependencies

if TYPE_CHECKING:
    from anndata import AnnData


def _parse_dataset_id(url: str):
    match = re.search(r"/e/(.+)", url)
    if match:
        return match.group(1).split(".")[0]
    else:
        raise ValueError(f"Could not parse dataset id from url {url}")


@dependencies("cellxgene_census")
def _load_cellxgene_dataset(
    url: str,
    filename: str | None = None,
    save_path: str = "data/",
    return_path: bool = False,
) -> AnnData | str:
    """Loads a file from `cellxgene <https://cellxgene.cziscience.com/>`_ portal.

    Parameters
    ----------
    url
        URL to cellxgene explorer
    collection_id
        Cellxgene collection ID. If None, will find the correct collection.
    filename
        manual override of the filename to write to.
    save_path
        Location to use when saving/loading the data.
    return_path
        If True, returns the path to the downloaded file instead of the AnnData object.

    Returns
    -------
    adata initialized with cellxgene data
    """
    from cellxgene_census import download_source_h5ad

    dataset_id = _parse_dataset_id(url)

    if filename is None:
        filename = f"{dataset_id}.h5ad"
    file_path = os.path.join(save_path, filename)
    # check if file exists
    if not os.path.exists(file_path):
        download_source_h5ad(dataset_id, to_path=file_path)
    if return_path:
        return file_path
    adata = read_h5ad(file_path)
    return adata



from __future__ import annotations

import os

import anndata
import numpy as np
import pytest
import torch
from torch.utils.data import Dataset

import scvi
from scvi import REGISTRY_KEYS
from tests.data.utils import generic_setup_adata_manager


def test_init():
    adata = scvi.data.synthetic_iid()
    manager = generic_setup_adata_manager(adata)

    dataset = manager.create_torch_dataset()
    assert dataset is not None
    assert isinstance(dataset, Dataset)
    assert len(dataset) > 0
    assert dataset.adata_manager is manager
    assert isinstance(dataset.keys_and_dtypes, dict)
    assert len(dataset.data) > 0
    assert len(dataset.registered_keys) > 0


def test_default_dtypes():
    adata = scvi.data.synthetic_iid()
    manager = generic_setup_adata_manager(adata, batch_key="batch")

    dataset = manager.create_torch_dataset()
    data = dataset[:10]
    assert isinstance(data, dict)
    assert REGISTRY_KEYS.X_KEY in data
    assert REGISTRY_KEYS.BATCH_KEY in data
    assert isinstance(data[REGISTRY_KEYS.X_KEY], np.ndarray)
    assert isinstance(data[REGISTRY_KEYS.BATCH_KEY], np.ndarray)
    assert data[REGISTRY_KEYS.X_KEY].dtype == np.float32
    assert data[REGISTRY_KEYS.BATCH_KEY].dtype == np.int64


def test_getitem_tensors():
    adata = scvi.data.synthetic_iid()
    manager = generic_setup_adata_manager(adata, batch_key="batch")

    # default
    dataset = manager.create_torch_dataset()
    assert isinstance(dataset.keys_and_dtypes, dict)
    assert REGISTRY_KEYS.X_KEY in dataset.keys_and_dtypes
    assert REGISTRY_KEYS.BATCH_KEY in dataset.keys_and_dtypes

    # list
    dataset = manager.create_torch_dataset(data_and_attributes=[REGISTRY_KEYS.X_KEY])
    assert isinstance(dataset.keys_and_dtypes, dict)
    assert list(dataset.keys_and_dtypes.keys()) == [REGISTRY_KEYS.X_KEY]

    # dict
    dataset = manager.create_torch_dataset(data_and_attributes={REGISTRY_KEYS.X_KEY: np.float64})
    assert isinstance(dataset.keys_and_dtypes, dict)
    assert list(dataset.keys_and_dtypes.keys()) == [REGISTRY_KEYS.X_KEY]
    assert dataset.keys_and_dtypes[REGISTRY_KEYS.X_KEY] == np.float64

    with pytest.raises(KeyError):
        manager.create_torch_dataset(data_and_attributes=["not_a_key"])

    with pytest.raises(KeyError):
        manager.create_torch_dataset(data_and_attributes=[REGISTRY_KEYS.CAT_COVS_KEY])

    with pytest.raises(ValueError):
        manager.create_torch_dataset(data_and_attributes=1)


def test_getitem(n_genes: int = 50):
    adata = scvi.data.synthetic_iid(n_genes=n_genes)
    manager = generic_setup_adata_manager(adata, batch_key="batch")

    dataset = manager.create_torch_dataset(load_sparse_tensor=True)

    data = dataset[:10]
    assert isinstance(data[REGISTRY_KEYS.X_KEY], np.ndarray)
    assert data[REGISTRY_KEYS.X_KEY].dtype == np.float32
    assert data[REGISTRY_KEYS.X_KEY].shape == (10, n_genes)
    assert isinstance(data[REGISTRY_KEYS.BATCH_KEY], np.ndarray)
    assert data[REGISTRY_KEYS.BATCH_KEY].dtype == np.int64
    assert data[REGISTRY_KEYS.BATCH_KEY].shape == (10, 1)

    data = dataset[[0, 1, 2]]
    assert isinstance(data[REGISTRY_KEYS.X_KEY], np.ndarray)
    assert data[REGISTRY_KEYS.X_KEY].dtype == np.float32
    assert data[REGISTRY_KEYS.X_KEY].shape == (3, n_genes)
    assert isinstance(data[REGISTRY_KEYS.BATCH_KEY], np.ndarray)
    assert data[REGISTRY_KEYS.BATCH_KEY].dtype == np.int64
    assert data[REGISTRY_KEYS.BATCH_KEY].shape == (3, 1)

    data = dataset[1]
    assert isinstance(data[REGISTRY_KEYS.X_KEY], np.ndarray)
    assert data[REGISTRY_KEYS.X_KEY].dtype == np.float32
    assert data[REGISTRY_KEYS.X_KEY].shape == (1, n_genes)
    assert isinstance(data[REGISTRY_KEYS.BATCH_KEY], np.ndarray)
    assert data[REGISTRY_KEYS.BATCH_KEY].dtype == np.int64
    assert data[REGISTRY_KEYS.BATCH_KEY].shape == (1, 1)


@pytest.mark.parametrize("sparse_format", ["csr_matrix", "csc_matrix"])
def test_load_sparse_tensor(sparse_format: str | None, n_genes: int = 50):
    adata = scvi.data.synthetic_iid(sparse_format=sparse_format, n_genes=n_genes)
    manager = generic_setup_adata_manager(adata, batch_key="batch")

    dataset = manager.create_torch_dataset(load_sparse_tensor=True)

    data = dataset[:10]
    assert isinstance(data[REGISTRY_KEYS.X_KEY], torch.Tensor)
    assert data[REGISTRY_KEYS.X_KEY].dtype == torch.float32
    assert data[REGISTRY_KEYS.X_KEY].shape == (10, n_genes)
    assert isinstance(data[REGISTRY_KEYS.BATCH_KEY], np.ndarray)
    assert data[REGISTRY_KEYS.BATCH_KEY].dtype == np.int64
    assert data[REGISTRY_KEYS.BATCH_KEY].shape == (10, 1)

    data = dataset[[0, 1, 2]]
    assert isinstance(data[REGISTRY_KEYS.X_KEY], torch.Tensor)
    assert data[REGISTRY_KEYS.X_KEY].dtype == torch.float32
    assert data[REGISTRY_KEYS.X_KEY].shape == (3, n_genes)
    assert isinstance(data[REGISTRY_KEYS.BATCH_KEY], np.ndarray)
    assert data[REGISTRY_KEYS.BATCH_KEY].dtype == np.int64
    assert data[REGISTRY_KEYS.BATCH_KEY].shape == (3, 1)

    data = dataset[1]
    assert isinstance(data[REGISTRY_KEYS.X_KEY], torch.Tensor)
    assert data[REGISTRY_KEYS.X_KEY].dtype == torch.float32
    assert data[REGISTRY_KEYS.X_KEY].shape == (1, n_genes)
    assert isinstance(data[REGISTRY_KEYS.BATCH_KEY], np.ndarray)
    assert data[REGISTRY_KEYS.BATCH_KEY].dtype == np.int64
    assert data[REGISTRY_KEYS.BATCH_KEY].shape == (1, 1)


def test_load_sparse_tensor_backed(save_path: str, n_genes: int = 50):
    adata = scvi.data.synthetic_iid(sparse_format="csr_matrix", n_genes=n_genes)
    adata_path = os.path.join(save_path, "adata.h5ad")
    adata.write(adata_path)
    del adata

    adata = anndata.read_h5ad(adata_path, backed="r")
    manager = generic_setup_adata_manager(adata, batch_key="batch")

    dataset = manager.create_torch_dataset(load_sparse_tensor=True)
    assert dataset.adata_manager.adata.isbacked

    data = dataset[:10]
    assert isinstance(data[REGISTRY_KEYS.X_KEY], torch.Tensor)
    assert data[REGISTRY_KEYS.X_KEY].dtype == torch.float32
    assert data[REGISTRY_KEYS.X_KEY].shape == (10, n_genes)
    assert isinstance(data[REGISTRY_KEYS.BATCH_KEY], np.ndarray)
    assert data[REGISTRY_KEYS.BATCH_KEY].dtype == np.int64
    assert data[REGISTRY_KEYS.BATCH_KEY].shape == (10, 1)

    data = dataset[[0, 1, 2]]
    assert isinstance(data[REGISTRY_KEYS.X_KEY], torch.Tensor)
    assert data[REGISTRY_KEYS.X_KEY].dtype == torch.float32
    assert data[REGISTRY_KEYS.X_KEY].shape == (3, n_genes)
    assert isinstance(data[REGISTRY_KEYS.BATCH_KEY], np.ndarray)
    assert data[REGISTRY_KEYS.BATCH_KEY].dtype == np.int64
    assert data[REGISTRY_KEYS.BATCH_KEY].shape == (3, 1)

    data = dataset[1]
    assert isinstance(data[REGISTRY_KEYS.X_KEY], torch.Tensor)
    assert data[REGISTRY_KEYS.X_KEY].dtype == torch.float32
    assert data[REGISTRY_KEYS.X_KEY].shape == (1, n_genes)
    assert isinstance(data[REGISTRY_KEYS.BATCH_KEY], np.ndarray)
    assert data[REGISTRY_KEYS.BATCH_KEY].dtype == np.int64
    assert data[REGISTRY_KEYS.BATCH_KEY].shape == (1, 1)



from __future__ import annotations

from math import ceil, floor

import numpy as np
import pytest

import scvi


@pytest.mark.parametrize("shuffle_set_split", [False, True])
def test_contrastive_datasplitter(
    shuffle_set_split: bool,
    mock_contrastive_adata_manager,
    mock_background_indices,
    mock_target_indices,
):
    background_indices = mock_background_indices
    target_indices = mock_target_indices

    contrastive_datasplitter = scvi.external.contrastivevi.ContrastiveDataSplitter(
        adata_manager=mock_contrastive_adata_manager,
        background_indices=background_indices,
        target_indices=target_indices,
        train_size=0.5,
        validation_size=0.3,
        shuffle_set_split=shuffle_set_split,
    )
    contrastive_datasplitter.setup()

    assert isinstance(contrastive_datasplitter.background_train_idx, list)
    assert isinstance(contrastive_datasplitter.background_val_idx, list)
    assert isinstance(contrastive_datasplitter.background_test_idx, list)
    assert isinstance(contrastive_datasplitter.target_train_idx, list)
    assert isinstance(contrastive_datasplitter.target_val_idx, list)
    assert isinstance(contrastive_datasplitter.target_test_idx, list)

    n_background = len(background_indices)
    n_background_train = ceil(n_background * 0.5)
    n_background_val = floor(n_background * 0.3)
    n_background_test = n_background - n_background_train - n_background_val

    assert len(contrastive_datasplitter.background_train_idx) == n_background_train
    assert len(contrastive_datasplitter.background_val_idx) == n_background_val
    assert len(contrastive_datasplitter.background_test_idx) == n_background_test

    n_target = len(target_indices)
    n_target_train = ceil(n_target * 0.5)
    n_target_val = floor(n_target * 0.3)
    n_target_test = n_target - n_target_train - n_target_val

    assert len(contrastive_datasplitter.target_train_idx) == n_target_train
    assert len(contrastive_datasplitter.target_val_idx) == n_target_val
    assert len(contrastive_datasplitter.target_test_idx) == n_target_test

    if not shuffle_set_split:
        np.testing.assert_array_equal(
            np.array(contrastive_datasplitter.background_val_idx),
            np.array(background_indices)[:n_background_val],
        )
        np.testing.assert_array_equal(
            np.array(contrastive_datasplitter.background_train_idx),
            np.array(background_indices)[
                n_background_val : (n_background_val + n_background_train)
            ],
        )
        np.testing.assert_array_equal(
            np.array(contrastive_datasplitter.background_test_idx),
            np.array(background_indices)[(n_background_val + n_background_train) :],
        )

        np.testing.assert_array_equal(
            np.array(contrastive_datasplitter.target_val_idx),
            np.array(target_indices)[:n_target_val],
        )
        np.testing.assert_array_equal(
            np.array(contrastive_datasplitter.target_train_idx),
            np.array(target_indices)[n_target_val : (n_target_val + n_target_train)],
        )
        np.testing.assert_array_equal(
            np.array(contrastive_datasplitter.target_test_idx),
            np.array(target_indices)[(n_target_val + n_target_train) :],
        )


@pytest.mark.parametrize("shuffle_set_split", [False, True])
def test_contrastive_datasplitter_with_external_indices(
    shuffle_set_split: bool,
    mock_contrastive_adata_manager,
    mock_background_indices,
    mock_target_indices,
):
    background_indices = mock_background_indices
    target_indices = mock_target_indices

    # define external indices apriori
    from sklearn.model_selection import train_test_split

    train_bg_ind, valid_bg_ind = train_test_split(background_indices, test_size=0.5)
    test_bg_ind, valid_bg_ind = train_test_split(valid_bg_ind, test_size=0.6)
    train_tr_ind, valid_tr_ind = train_test_split(target_indices, test_size=0.495)
    test_tr_ind, valid_tr_ind = train_test_split(valid_tr_ind, test_size=0.6)

    contrastive_datasplitter = scvi.external.contrastivevi.ContrastiveDataSplitter(
        adata_manager=mock_contrastive_adata_manager,
        background_indices=background_indices,
        target_indices=target_indices,
        train_size=0.5,
        validation_size=0.3,
        shuffle_set_split=shuffle_set_split,
        external_indexing=[
            np.array(train_bg_ind + train_tr_ind),
            np.array(valid_bg_ind + valid_tr_ind),
            np.array(test_bg_ind + test_tr_ind),
        ],
    )
    contrastive_datasplitter.setup()

    assert isinstance(contrastive_datasplitter.background_train_idx, list)
    assert isinstance(contrastive_datasplitter.background_val_idx, list)
    assert isinstance(contrastive_datasplitter.background_test_idx, list)
    assert isinstance(contrastive_datasplitter.target_train_idx, list)
    assert isinstance(contrastive_datasplitter.target_val_idx, list)
    assert isinstance(contrastive_datasplitter.target_test_idx, list)

    n_background = len(background_indices)
    n_background_train = ceil(n_background * 0.5)
    n_background_val = floor(n_background * 0.3)
    n_background_test = n_background - n_background_train - n_background_val

    assert len(contrastive_datasplitter.background_train_idx) == n_background_train
    assert len(contrastive_datasplitter.background_val_idx) == n_background_val
    assert len(contrastive_datasplitter.background_test_idx) == n_background_test

    n_target = len(target_indices)
    n_target_train = ceil(n_target * 0.5)
    n_target_val = floor(n_target * 0.3)
    n_target_test = n_target - n_target_train - n_target_val

    assert len(contrastive_datasplitter.target_train_idx) == n_target_train
    assert len(contrastive_datasplitter.target_val_idx) == n_target_val
    assert len(contrastive_datasplitter.target_test_idx) == n_target_test

    # order of indexing is not matter with external indices so we will pass this check here



import os

import anndata as ad
import numpy as np
import pytest
import scanpy as sc
from mudata import MuData

import scvi
from scvi import REGISTRY_KEYS
from scvi.data import synthetic_iid
from scvi.model import MULTIVI
from scvi.utils import attrdict


def test_multivi():
    data = synthetic_iid()
    MULTIVI.setup_anndata(
        data,
        batch_key="batch",
    )
    vae = MULTIVI(
        data,
        n_genes=50,
        n_regions=50,
    )
    vae.train(1, save_best=False)
    vae.train(1, adversarial_mixing=False)
    vae.train(3)
    vae.get_elbo(indices=vae.validation_indices)
    vae.get_accessibility_estimates()
    vae.get_accessibility_estimates(normalize_cells=True)
    vae.get_accessibility_estimates(normalize_regions=True)
    vae.get_normalized_expression()
    vae.get_library_size_factors()
    vae.get_region_factors()
    vae.get_reconstruction_error(indices=vae.validation_indices)
    vae.get_latent_representation()
    vae.differential_accessibility(groupby="labels", group1="label_1")
    vae.differential_expression(groupby="labels", group1="label_1")

    # Test with size factor
    data = synthetic_iid()
    data.obs["size_factor"] = np.random.randint(1, 5, size=(data.shape[0],))
    MULTIVI.setup_anndata(data, batch_key="batch")
    vae = MULTIVI(
        data,
        n_genes=50,
        n_regions=50,
    )
    vae.train(3)

    # Test with modality weights and penalties
    data = synthetic_iid()
    MULTIVI.setup_anndata(data, batch_key="batch")
    vae = MULTIVI(data, n_genes=50, n_regions=50, modality_weights="cell")
    vae.train(3)
    vae = MULTIVI(data, n_genes=50, n_regions=50, modality_weights="universal")
    vae.train(3)
    vae = MULTIVI(data, n_genes=50, n_regions=50, modality_penalty="MMD")
    vae.train(3)

    # Test with non-zero protein data
    data = synthetic_iid()
    MULTIVI.setup_anndata(
        data,
        batch_key="batch",
        protein_expression_obsm_key="protein_expression",
        protein_names_uns_key="protein_names",
    )
    vae = MULTIVI(
        data,
        n_genes=50,
        n_regions=50,
        modality_weights="cell",
    )
    assert vae.n_proteins == data.obsm["protein_expression"].shape[1]
    vae.train(3)


def test_multivi_single_batch():
    data = synthetic_iid(n_batches=1)
    MULTIVI.setup_anndata(data, batch_key="batch")
    vae = MULTIVI(data, n_genes=50, n_regions=50)
    with pytest.warns(UserWarning):
        vae.train(3)


@pytest.mark.internet
def test_multivi_mudata_rna_prot_external():
    # Example on how to download protein adata to mudata (from multivi tutorial) - mudata RNA/PROT
    adata = scvi.data.pbmcs_10x_cite_seq()
    adata.layers["counts"] = adata.X.copy()
    adata.obs_names_make_unique()
    protein_adata = ad.AnnData(adata.obsm["protein_expression"])
    protein_adata.obs_names = adata.obs_names
    del adata.obsm["protein_expression"]
    mdata = MuData({"rna": adata, "protein": protein_adata})
    sc.pp.highly_variable_genes(
        mdata.mod["rna"],
        n_top_genes=4000,
        flavor="seurat_v3",
        batch_key="batch",
        layer="counts",
    )
    mdata.mod["rna_subset"] = mdata.mod["rna"][:, mdata.mod["rna"].var["highly_variable"]].copy()
    mdata.update()
    # mdata
    # mdata.mod
    MULTIVI.setup_mudata(
        mdata,
        rna_layer="counts",  # mean we use: mdata.mod["rna_subset"].layers["counts"]
        protein_layer=None,  # mean we use: mdata.mod["protein"].X
        batch_key="batch",  # the batch is here: mdata.mod["rna_subset"].obs["batch"]
        modalities={
            "rna_layer": "rna_subset",
            "protein_layer": "protein",
            "batch_key": "rna_subset",
        },
    )
    model = MULTIVI(mdata)
    model.train(1, train_size=0.9)


def test_multivi_mudata_rna_atac_external():
    # optional data - mudata RNA/ATAC
    mdata = synthetic_iid(return_mudata=True)
    sc.pp.highly_variable_genes(
        mdata.mod["rna"],
        n_top_genes=4000,
        flavor="seurat_v3",
    )
    mdata.mod["rna_subset"] = mdata.mod["rna"][:, mdata.mod["rna"].var["highly_variable"]].copy()
    sc.pp.highly_variable_genes(
        mdata.mod["accessibility"],
        n_top_genes=4000,
        flavor="seurat_v3",
    )
    mdata.mod["atac_subset"] = mdata.mod["accessibility"][
        :, mdata.mod["accessibility"].var["highly_variable"]
    ].copy()
    mdata.update()
    MULTIVI.setup_mudata(
        mdata,
        modalities={"rna_layer": "rna_subset", "atac_layer": "atac_subset"},
    )
    model = MULTIVI(mdata)
    model.train(1, train_size=0.9)


def test_multivi_mudata_trimodal_external():
    # optional data - mudata RNA/ATAC
    mdata = synthetic_iid(return_mudata=True)
    MULTIVI.setup_mudata(
        mdata,
        modalities={
            "rna_layer": "rna",
            "atac_layer": "accessibility",
            "protein_layer": "protein_expression",
        },
    )
    model = MULTIVI(mdata)
    model.train(1, train_size=0.9)
    model.train(1, train_size=0.9)
    assert model.is_trained is True
    model.get_latent_representation()
    model.get_elbo()
    model.get_reconstruction_error()
    model.get_normalized_expression()
    model.get_accessibility_estimates()
    model.get_accessibility_estimates(normalize_cells=True)
    model.get_accessibility_estimates(normalize_regions=True)
    model.get_library_size_factors()
    model.get_region_factors()

    model.get_elbo(indices=model.validation_indices)
    model.get_reconstruction_error(indices=model.validation_indices)
    model.get_accessibility_estimates()
    model.get_accessibility_estimates(normalize_cells=True)
    model.get_accessibility_estimates(normalize_regions=True)
    model.get_library_size_factors()
    model.get_region_factors()


@pytest.mark.parametrize("n_genes", [25, 50, 100])
@pytest.mark.parametrize("n_regions", [25, 50, 100])
def test_multivi_mudata(n_genes: int, n_regions: int):
    # use of syntetic data of rna/proteins/atac for speed

    mdata = synthetic_iid(return_mudata=True)
    MULTIVI.setup_mudata(
        mdata,
        batch_key="batch",
        modalities={
            "rna_layer": "rna",
            "protein_layer": "protein_expression",
            "atac_layer": "accessibility",
        },
    )
    n_obs = mdata.n_obs
    n_latent = 10

    model = MULTIVI(mdata, n_latent=n_latent, n_genes=n_genes, n_regions=n_regions)
    model.train(1, train_size=0.9)
    assert model.is_trained is True
    z = model.get_latent_representation()
    assert z.shape == (n_obs, n_latent)
    model.get_elbo()
    model.get_reconstruction_error()
    model.get_normalized_expression()
    model.get_normalized_expression(transform_batch=["batch_0", "batch_1"])
    model.get_accessibility_estimates()
    model.get_accessibility_estimates(normalize_cells=True)
    model.get_accessibility_estimates(normalize_regions=True)
    model.get_library_size_factors()
    model.get_region_factors()

    model.get_elbo(indices=model.validation_indices)
    model.get_reconstruction_error(indices=model.validation_indices)
    model.get_accessibility_estimates()
    model.get_accessibility_estimates(normalize_cells=True)
    model.get_accessibility_estimates(normalize_regions=True)
    model.get_library_size_factors()
    model.get_region_factors()

    mdata2 = synthetic_iid(return_mudata=True)
    MULTIVI.setup_mudata(
        mdata2,
        batch_key="batch",
        modalities={"rna_layer": "rna", "protein_layer": "protein_expression"},
    )
    norm_exp = model.get_normalized_expression(mdata2, indices=[1, 2, 3])
    assert norm_exp.shape == (3, n_genes)

    # test transfer_anndata_setup + view
    mdata3 = synthetic_iid(return_mudata=True)
    mdata3.obs["_indices"] = np.arange(mdata3.n_obs)
    model.get_elbo(mdata3[:10])
    model.get_accessibility_estimates()
    model.get_accessibility_estimates(normalize_cells=True)
    model.get_accessibility_estimates(normalize_regions=True)
    model.get_library_size_factors()
    model.get_region_factors()


def test_multivi_auto_transfer_mudata():
    # test automatic transfer_fields
    adata = synthetic_iid()
    protein_adata = synthetic_iid(n_genes=50)
    mdata = MuData({"rna": adata, "protein": protein_adata})
    MULTIVI.setup_mudata(
        mdata,
        batch_key="batch",
        modalities={"rna_layer": "rna", "batch_key": "rna", "protein_layer": "protein"},
    )
    model = MULTIVI(mdata)
    adata2 = synthetic_iid()
    protein_adata2 = synthetic_iid(n_genes=50)
    mdata2 = MuData({"rna": adata2, "protein": protein_adata2})
    mdata2.obs["_indices"] = np.arange(mdata2.n_obs)
    model.get_elbo(mdata2)
    model.get_accessibility_estimates()
    model.get_accessibility_estimates(normalize_cells=True)
    model.get_accessibility_estimates(normalize_regions=True)
    model.get_library_size_factors()
    model.get_region_factors()


def test_multivi_incorrect_mapping_mudata():
    # test that we catch incorrect mappings
    adata = synthetic_iid()
    protein_adata = synthetic_iid(n_genes=50)
    mdata = MuData({"rna": adata, "protein": protein_adata})
    MULTIVI.setup_mudata(
        mdata,
        batch_key="batch",
        modalities={"rna_layer": "rna", "batch_key": "rna", "protein_layer": "protein"},
    )
    model = MULTIVI(mdata)
    adata2 = synthetic_iid()
    protein_adata2 = synthetic_iid(n_genes=50)
    mdata2 = MuData({"rna": adata2, "protein": protein_adata2})
    adata2.obs.batch = adata2.obs.batch.cat.rename_categories(["batch_0", "batch_10"])
    with pytest.raises(ValueError):
        model.get_elbo(mdata2)


def test_multivi_reordered_mapping_mudata():
    # test that same mapping different order is okay
    adata = synthetic_iid()
    protein_adata = synthetic_iid(n_genes=50)
    mdata = MuData({"rna": adata, "protein": protein_adata})
    MULTIVI.setup_mudata(
        mdata,
        batch_key="batch",
        modalities={"rna_layer": "rna", "batch_key": "rna", "protein_layer": "protein"},
    )
    model = MULTIVI(mdata)
    adata2 = synthetic_iid()
    protein_adata2 = synthetic_iid(n_genes=50)
    mdata2 = MuData({"rna": adata2, "protein": protein_adata2})
    adata2.obs.batch = adata2.obs.batch.cat.rename_categories(["batch_1", "batch_0"])
    mdata2.obs["_indices"] = np.arange(mdata2.n_obs)
    model.get_elbo(mdata2)
    model.get_accessibility_estimates()
    model.get_accessibility_estimates(normalize_cells=True)
    model.get_accessibility_estimates(normalize_regions=True)
    model.get_library_size_factors()
    model.get_region_factors()


def test_multivi_model_library_size_mudata():
    adata = synthetic_iid()
    protein_adata = synthetic_iid(n_genes=50)
    mdata = MuData({"rna": adata, "protein": protein_adata})
    MULTIVI.setup_mudata(
        mdata,
        batch_key="batch",
        modalities={"rna_layer": "rna", "batch_key": "rna", "protein_layer": "protein"},
    )

    n_latent = 10
    model = MULTIVI(mdata, n_latent=n_latent)
    model.train(1, train_size=0.5)
    assert model.is_trained is True
    model.get_elbo()
    model.get_accessibility_estimates()
    model.get_accessibility_estimates(normalize_cells=True)
    model.get_accessibility_estimates(normalize_regions=True)
    model.get_library_size_factors()
    model.get_region_factors()


def test_multivi_size_factor_mudata():
    mdata = synthetic_iid(return_mudata=True)
    mdata.obs["size_factor_rna"] = mdata["rna"].X.sum(1)
    mdata.obs["size_factor_atac"] = (mdata["accessibility"].X.sum(1) + 1) / (
        np.max(mdata["accessibility"].X.sum(1)) + 1.01
    )
    MULTIVI.setup_mudata(
        mdata,
        modalities={"rna_layer": "rna", "atac_layer": "accessibility"},
        size_factor_key=["size_factor_rna", "size_factor_atac"],
    )

    n_latent = 10

    # Test size_factor_key overrides use_observed_lib_size.
    model = MULTIVI(mdata, n_latent=n_latent)
    assert model.module.use_size_factor_key
    model.train(1, train_size=0.5)

    model = MULTIVI(mdata, n_latent=n_latent)
    assert model.module.use_size_factor_key
    model.train(1, train_size=0.5)


def test_multivi_saving_and_loading_mudata(save_path: str):
    adata = synthetic_iid()
    protein_adata = synthetic_iid(n_genes=50)
    mdata = MuData({"rna": adata, "protein": protein_adata})
    MULTIVI.setup_mudata(
        mdata,
        batch_key="batch",
        modalities={"rna_layer": "rna", "batch_key": "rna", "protein_layer": "protein"},
    )
    model = MULTIVI(mdata)
    model.train(1, train_size=0.2)
    z1 = model.get_latent_representation(mdata)
    test_idx1 = model.validation_indices

    model.save(save_path, overwrite=True, save_anndata=True)
    model.view_setup_args(save_path)

    model = MULTIVI.load(save_path)
    model.get_latent_representation()

    # Load with mismatched genes.
    tmp_adata = synthetic_iid(
        n_genes=200,
    )
    tmp_protein_adata = synthetic_iid(n_genes=50)
    tmp_mdata = MuData({"rna": tmp_adata, "protein": tmp_protein_adata})
    with pytest.raises(ValueError):
        MULTIVI.load(save_path, adata=tmp_mdata)

    # Load with different batches.
    tmp_adata = synthetic_iid()
    tmp_adata.obs["batch"] = tmp_adata.obs["batch"].cat.rename_categories(["batch_2", "batch_3"])
    tmp_protein_adata = synthetic_iid(n_genes=50)
    tmp_mdata = MuData({"rna": tmp_adata, "protein": tmp_protein_adata})
    with pytest.raises(ValueError):
        MULTIVI.load(save_path, adata=tmp_mdata)

    model = MULTIVI.load(save_path, adata=mdata)
    assert REGISTRY_KEYS.BATCH_KEY in model.adata_manager.data_registry
    assert model.adata_manager.data_registry.batch == attrdict(
        {"mod_key": "rna", "attr_name": "obs", "attr_key": "_scvi_batch"}
    )

    z2 = model.get_latent_representation()
    test_idx2 = model.validation_indices
    np.testing.assert_array_equal(z1, z2)
    np.testing.assert_array_equal(test_idx1, test_idx2)
    assert model.is_trained is True

    save_path = os.path.join(save_path, "tmp")

    adata2 = synthetic_iid()
    protein_adata2 = synthetic_iid(n_genes=50)
    mdata2 = MuData({"rna": adata2, "protein": protein_adata2})
    MULTIVI.setup_mudata(
        mdata2,
        batch_key="batch",
        modalities={"rna_layer": "rna", "batch_key": "rna", "protein_layer": "protein"},
    )


def test_scarches_mudata_prep_layer(save_path: str):
    n_latent = 5
    mdata1 = synthetic_iid(return_mudata=True)

    mdata1["rna"].layers["counts"] = mdata1["rna"].X.copy()
    MULTIVI.setup_mudata(
        mdata1,
        batch_key="batch",
        modalities={"rna_layer": "rna", "protein_layer": "protein_expression"},
    )
    model = MULTIVI(mdata1, n_latent=n_latent)
    model.train(1, check_val_every_n_epoch=1)
    dir_path = os.path.join(save_path, "saved_model/")
    model.save(dir_path, overwrite=True)

    # mdata2 has more genes and missing 10 genes from mdata1.
    # protein/acessibility features are same as in mdata1
    mdata2 = synthetic_iid(n_genes=110, return_mudata=True)
    mdata2["rna"].layers["counts"] = mdata2["rna"].X.copy()
    new_var_names_init = [f"Random {i}" for i in range(10)]
    new_var_names = new_var_names_init + mdata2["rna"].var_names[10:].to_list()
    mdata2["rna"].var_names = new_var_names

    original_protein_values = mdata2["protein_expression"].X.copy()
    original_accessibility_values = mdata2["accessibility"].X.copy()

    MULTIVI.prepare_query_mudata(mdata2, dir_path)
    # should be padded 0s
    assert np.sum(mdata2["rna"][:, mdata2["rna"].var_names[:10]].layers["counts"]) == 0
    np.testing.assert_equal(
        mdata2["rna"].var_names[:10].to_numpy(), mdata1["rna"].var_names[:10].to_numpy()
    )

    # values of other modalities should be unchanged
    np.testing.assert_equal(original_protein_values, mdata2["protein_expression"].X)
    np.testing.assert_equal(original_accessibility_values, mdata2["accessibility"].X)

    # and names should also be the same
    np.testing.assert_equal(
        mdata2["protein_expression"].var_names.to_numpy(),
        mdata1["protein_expression"].var_names.to_numpy(),
    )
    np.testing.assert_equal(
        mdata2["accessibility"].var_names.to_numpy(), mdata1["accessibility"].var_names.to_numpy()
    )
    MULTIVI.load_query_data(mdata2, dir_path)


def test_multivi_save_load_mudata_format(save_path: str):
    mdata = synthetic_iid(return_mudata=True, protein_expression_key="protein")
    invalid_mdata = mdata.copy()
    invalid_mdata.mod["protein"] = invalid_mdata.mod["protein"][:, :10].copy()
    MULTIVI.setup_mudata(
        mdata,
        modalities={"rna_layer": "rna", "protein_layer": "protein"},
    )
    model = MULTIVI(mdata)
    model.train(max_epochs=1)

    legacy_model_path = os.path.join(save_path, "legacy_model")
    model.save(
        legacy_model_path,
        overwrite=True,
        save_anndata=False,
        legacy_mudata_format=True,
    )

    with pytest.raises(ValueError):
        _ = MULTIVI.load(legacy_model_path, adata=invalid_mdata)
    model = MULTIVI.load(legacy_model_path, adata=mdata)

    model_path = os.path.join(save_path, "model")
    model.save(
        model_path,
        overwrite=True,
        save_anndata=False,
        legacy_mudata_format=False,
    )
    with pytest.raises(ValueError):
        _ = MULTIVI.load(legacy_model_path, adata=invalid_mdata)
    model = MULTIVI.load(model_path, adata=mdata)



from ._beta_binomial import BetaBinomial
from ._negative_binomial import (
    JaxNegativeBinomialMeanDisp,
    NegativeBinomial,
    NegativeBinomialMixture,
    Poisson,
    ZeroInflatedNegativeBinomial,
)
from ._normal import Normal

__all__ = [
    "NegativeBinomial",
    "NegativeBinomialMixture",
    "ZeroInflatedNegativeBinomial",
    "JaxNegativeBinomialMeanDisp",
    "Poisson",
    "BetaBinomial",
    "Normal",
]
